"""Celery tasks for Member Audit."""

# pylint: disable=redefined-builtin, too-many-lines

import inspect
import random
from typing import Callable, Iterable, List, Optional

from celery import chain, shared_task

from django.apps import apps
from django.contrib.auth.models import Group, User
from esi.models import Token
from eveuniverse.constants import POST_UNIVERSE_NAMES_MAX_ITEMS
from eveuniverse.models import EveEntity, EveMarketPrice

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.esi import EsiErrorLimitExceeded, EsiOffline
from app_utils.logging import LoggerAddTag

from memberaudit import __title__, utils
from memberaudit.app_settings import (
    MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
    MEMBERAUDIT_SECTION_STALE_MINUTES_GLOBAL_DEFAULT,
    MEMBERAUDIT_TASKS_LOW_PRIORITY,
    MEMBERAUDIT_TASKS_MAX_ASSETS_PER_PASS,
    MEMBERAUDIT_TASKS_NORMAL_PRIORITY,
    MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT,
    MEMBERAUDIT_TASKS_TIME_LIMIT,
)
from memberaudit.core import data_exporters, esi_status
from memberaudit.decorators import when_esi_is_available
from memberaudit.helpers import determine_task_priority
from memberaudit.models import (
    Character,
    CharacterAsset,
    CharacterContract,
    ComplianceGroupDesignation,
    General,
    Location,
    MailEntity,
    enabled_sections_by_stale_minutes,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

MAX_RETRIES_DEFAULT = 3
MAX_RETRIES_STRUCTURES = 5
MAX_RETRIES_MAIL_ENTITIES = 5

TASK_DEFAULTS = {
    "time_limit": MEMBERAUDIT_TASKS_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}
"""Default params for all tasks."""

TASK_DEFAULTS_BIND = {**TASK_DEFAULTS, **{"bind": True}}
"""Default params for tasks that need access to self."""

TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}
"""Default params for tasks that need run once only."""

TASK_DEFAULTS_BIND_ONCE = {**TASK_DEFAULTS, **{"bind": True, "base": QueueOnce}}
"""Default params for tasks that need access to self and run once only."""


@shared_task(**TASK_DEFAULTS_ONCE)
def run_regular_updates() -> None:
    """Run regular updates for Member Audit."""
    update_market_prices.apply_async(priority=MEMBERAUDIT_TASKS_LOW_PRIORITY)
    update_all_characters.apply_async(priority=MEMBERAUDIT_TASKS_LOW_PRIORITY)
    if ComplianceGroupDesignation.objects.exists():
        update_compliance_groups_for_all.apply_async(
            priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
@when_esi_is_available
def update_all_characters(
    self, force_update: bool = False, ignore_stale: bool = False
) -> None:
    """Update all enabled characters from ESI and disable update for orphans.

    Args:
    - force_update: When set to True will always refresh from data from ESI,
        even if it has not changed
    - ignore_stale: When True, will start updating all sections regardless
        of it's stale status
    """
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY

    # check consistency of shared characters
    shared_characters = list(
        Character.objects.filter(is_shared=True).values_list("pk", flat=True)
    )
    if shared_characters:
        for character_pk in shared_characters:
            check_character_consistency.apply_async(
                kwargs={"character_pk": character_pk}, priority=priority
            )

    Character.objects.disable_characters_with_no_owner()

    # start sync for all enabled characters
    enabled_characters = list(
        Character.objects.filter(is_disabled=False).values_list("pk", flat=True)
    )
    if enabled_characters:
        for character_pk in enabled_characters:
            update_character.apply_async(
                kwargs={
                    "character_pk": character_pk,
                    "force_update": force_update,
                    "ignore_stale": ignore_stale,
                },
                priority=priority,
            )
        logger.info("Started updates for %d characters", len(enabled_characters))

    else:
        logger.info("No enabled characters found for update.")


# Main character update tasks


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
def update_character(
    self, character_pk: int, force_update: bool = False, ignore_stale: bool = False
) -> bool:
    """Update all sections of a character from ESI.

    Args:
    - character_pk: PL of character to update
    - force_update: When set to True will always refresh from data from ESI,
        even if it has not changed
    - ignore_stale: When True, will start updating all sections regardless
        of it's stale status

    Return True when update of sections was started, else False.
    """
    character: Character = Character.objects.prefetch_related("update_status_set").get(
        pk=character_pk
    )
    if character.is_orphan:
        logger.info("%s: Skipping update for orphaned character", character)
        return False

    character.reset_token_error_notified_if_status_ok()
    character.clear_cache()

    character_needs_update = character.calc_update_needed()
    if not ignore_stale and not character_needs_update:
        logger.info("%s: No update required", character)
        return False

    msg = f"{character}: Starting character update ({force_update=}, {ignore_stale=})"
    logger.info(msg)
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY

    unavailable_sections = esi_status.unavailable_sections()
    if unavailable_sections is None:
        logger.warning(
            "%s: Failed to determine from ESI which sections are available. Update aborted.",
            character,
        )
        return False

    for section in enabled_sections_by_stale_minutes():
        if section in unavailable_sections:
            logger.warning(
                "%s: Skipping update for this section because ESI reports it as unavailable: %s",
                character,
                section,
            )
            continue

        if not ignore_stale and not character_needs_update.for_section(section):
            continue

        task_name = f"update_character_{section.value}"
        task = globals()[task_name]
        task.apply_async(
            kwargs={"character_pk": character.pk, "force_update": force_update},
            priority=priority,
        )

    return True


# Updating sections with simple update logic


_update_character_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["character_pk", "force_update"], "graceful": True}},
}


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_attributes(character_pk: int, force_update: bool) -> None:
    """Update attributes for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.ATTRIBUTES,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_character_details(character_pk: int, force_update: bool) -> None:
    """Update character_details for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.CHARACTER_DETAILS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_corporation_history(character_pk: int, force_update: bool) -> None:
    """Update corporation_history for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.CORPORATION_HISTORY,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_fw_stats(character_pk: int, force_update: bool) -> None:
    """Update fw_stats for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.FW_STATS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_implants(character_pk: int, force_update: bool) -> None:
    """Update implants for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.IMPLANTS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_jump_clones(character_pk: int, force_update: bool) -> None:
    """Update jump_clones for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.JUMP_CLONES,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_location(character_pk: int, force_update: bool) -> None:
    """Update location for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.LOCATION,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_loyalty(character_pk: int, force_update: bool) -> None:
    """Update loyalty for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.LOYALTY,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_mining_ledger(character_pk: int, force_update: bool) -> None:
    """Update mining_ledger for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.MINING_LEDGER,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_online_status(character_pk: int, force_update: bool) -> None:
    """Update online_status for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.ONLINE_STATUS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_planets(character_pk: int, force_update: bool) -> None:
    """Update planets for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.PLANETS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_roles(character_pk: int, force_update: bool) -> None:
    """Update roles for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.ROLES,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_ship(character_pk: int, force_update: bool) -> None:
    """Update ship for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.SHIP,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_skill_queue(character_pk: int, force_update: bool) -> None:
    """Update skill_queue for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.SKILL_QUEUE,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_skill_sets(character_pk: int, force_update: bool) -> None:
    """Update skill_sets for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.SKILL_SETS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_skills(character_pk: int, force_update: bool) -> None:
    """Update skills for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.SKILLS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_standings(character_pk: int, force_update: bool) -> None:
    """Update standings for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.STANDINGS,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_titles(character_pk: int, force_update: bool) -> None:
    """Update titles for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.TITLES,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_wallet_balance(character_pk: int, force_update: bool) -> None:
    """Update wallet_balance for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.WALLET_BALLANCE,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_wallet_journal(character_pk: int, force_update: bool) -> None:
    """Update wallet_journal for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.WALLET_JOURNAL,
        force_update=force_update,
    )


@shared_task(**_update_character_params)
@when_esi_is_available
def update_character_wallet_transactions(character_pk: int, force_update: bool) -> None:
    """Update wallet_transactions for a character from ESI."""
    _update_character_section(
        character_pk=character_pk,
        section=Character.UpdateSection.WALLET_TRANSACTIONS,
        force_update=force_update,
    )


def _update_character_section(
    character_pk: int, section: str, force_update: bool
) -> None:
    """Update a section for a character from ESI."""
    section = Character.UpdateSection(section)
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    character.reset_update_section(section)
    logger.info("%s: Updating %s", character, section.label)

    method: Callable = getattr(character, section.method_name)
    method_signature = inspect.signature(method)
    if "force_update" in method_signature.parameters:
        kwargs = {"force_update": force_update}
    else:
        kwargs = {}

    result = character.perform_update_with_error_logging(
        section=section, method=method, **kwargs
    )
    character.update_section_log_result(
        section, is_success=True, is_updated=result.is_updated
    )


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_unresolved_eve_entities() -> None:
    """Bulk resolved all unresolved EveEntity objects in database.

    This task is used by other apps. Do not remove!
    """
    unresolved_ids = EveEntity.objects.filter(name="")[
        :POST_UNIVERSE_NAMES_MAX_ITEMS
    ].values_list("id", flat=True)
    if unresolved_ids:
        updated_count = EveEntity.objects.update_from_esi_by_id(unresolved_ids)
        logger.info("Updating %d unresolved entities from ESI", updated_count)


# Special tasks for updating assets


_update_task_params_with_bind = {
    **TASK_DEFAULTS_BIND_ONCE,
    **{"once": {"keys": ["character_pk", "force_update"], "graceful": True}},
}


@shared_task(**_update_task_params_with_bind)
def update_character_assets(self, character_pk: int, force_update: bool) -> None:
    """Update the assets of a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    section = Character.UpdateSection.ASSETS
    logger.info("%s: Updating %s", character, section.label)
    character.reset_update_section(section=section)
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    chain(
        assets_build_list_from_esi.s(character.pk, force_update).set(priority=priority),
        assets_preload_objects.s(character.pk).set(priority=priority),
        assets_create_parents.s(character.pk).set(priority=priority),
    ).delay()


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def assets_build_list_from_esi(
    character_pk: int, force_update: bool = False
) -> Optional[dict]:
    """Retrieve asset list for a character from ESI and return it
    or return None if asset list is unchanged.
    """
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    result = character.perform_update_with_error_logging(
        section=Character.UpdateSection.ASSETS,
        method=character.assets_build_list_from_esi,
        force_update=force_update,
    )
    if not result.is_changed and not force_update:
        return None

    asset_list = result.data
    if asset_list:
        ship_asset_record = character.generate_asset_from_current_ship_and_location()
        if ship_asset_record:
            ship_item_id = ship_asset_record["item_id"]
            asset_item_ids = {item["item_id"] for item in asset_list}
            if ship_item_id not in asset_item_ids:
                asset_list.append(ship_asset_record)
                logger.info("%s: Added current ship to assets", character)

    return asset_list


@shared_task(**TASK_DEFAULTS)
def assets_preload_objects(
    asset_list: Optional[list], character_pk: int
) -> Optional[list]:
    """Preload asset objects for a character from ESI."""
    if asset_list is None:
        return None  # Exit when assets are unchanged

    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    character.perform_update_with_error_logging(
        Character.UpdateSection.ASSETS,
        character.assets_preload_objects,
        asset_list,
    )
    return asset_list


@shared_task(**TASK_DEFAULTS_BIND)
def assets_create_parents(self, asset_list: Optional[list], character_pk: int) -> None:
    """Create the parent assets from an asset list.

    Parent assets are assets attached directly to a Location object (e.g. station)

    This task will recursively call itself until all possible parent assets
    from the asset list have been created.
    Then call another task to create child assets.
    """
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    if asset_list is None:
        character.update_section_log_result(
            Character.UpdateSection.ASSETS, is_success=True, is_updated=False
        )
        return

    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY

    asset_data_initial = {asset["item_id"]: asset for asset in asset_list}
    asset_data = _assets_create_parents_chunk(
        character=character, asset_data=asset_data_initial, cycle=1
    )

    if asset_data:
        assets_create_children.apply_async(
            kwargs={
                "asset_list": list(asset_data.values()),
                "character_pk": character.pk,
            },
            priority=priority,
        )
    else:
        character.update_section_log_result(
            Character.UpdateSection.ASSETS, is_success=True, is_updated=True
        )


def _assets_create_parents_chunk(character: Character, asset_data: dict, cycle: int):
    """Create chunk of parent assets for a character."""
    logger.info("%s: Creating parent assets - pass %s", character, cycle)
    new_assets = []
    if cycle == 1:
        character.assets.all().delete()

    known_location_ids = set(Location.objects.values_list("id", flat=True))
    parent_asset_ids = {
        item_id
        for item_id, asset_info in asset_data.items()
        if asset_info["location_id"] in known_location_ids
    }
    for item_id in parent_asset_ids:
        item = asset_data[item_id]
        new_assets.append(
            CharacterAsset(
                character=character,
                item_id=item_id,
                location_id=item["location_id"],
                eve_type_id=item["type_id"],
                name=item["name"],
                is_blueprint_copy=item.get("is_blueprint_copy"),
                is_singleton=item["is_singleton"],
                location_flag=item["location_flag"],
                quantity=item["quantity"],
            )
        )
        asset_data.pop(item_id)
        if len(new_assets) >= MEMBERAUDIT_TASKS_MAX_ASSETS_PER_PASS:
            break

    created_objs = CharacterAsset.objects.bulk_create_with_fallback(
        new_assets, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
    )
    logger.info("%s: Created %s parent assets", character, len(created_objs))

    if len(parent_asset_ids) > len(new_assets):
        # there are more parent assets to create
        asset_data = _assets_create_parents_chunk(
            character=character, asset_data=asset_data, cycle=cycle + 1
        )

    return asset_data


@shared_task(**TASK_DEFAULTS_BIND)
def assets_create_children(
    self, asset_list: list, character_pk: int, cycle: int = 1
) -> None:
    """Create child assets from given asset list.

    Child assets are assets located within other assets (aka containers)

    This task will recursively call itself until all possible assets from the
    asset list are included into the asset tree
    """
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    logger.info("%s: Creating child assets - pass %s", character, cycle)

    # for debug
    # store_list_to_disk(character, asset_data, f"child_asset_list_{cycle}")

    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    asset_data = {asset["item_id"]: asset for asset in asset_list}
    new_assets = []
    parent_asset_ids = set(character.assets.values_list("item_id", flat=True))
    child_asset_ids = {
        item_id
        for item_id, item in asset_data.items()
        if item["location_id"] in parent_asset_ids
    }
    for item_id in child_asset_ids:
        item = asset_data[item_id]
        new_assets.append(
            CharacterAsset(
                character=character,
                item_id=item_id,
                parent=character.assets.get(item_id=item["location_id"]),
                eve_type_id=item["type_id"],
                name=item["name"],
                is_blueprint_copy=item.get("is_blueprint_copy"),
                is_singleton=item["is_singleton"],
                location_flag=item["location_flag"],
                quantity=item["quantity"],
            )
        )
        asset_data.pop(item_id)
        if len(new_assets) >= MEMBERAUDIT_TASKS_MAX_ASSETS_PER_PASS:
            break

    if new_assets:
        created_objs = CharacterAsset.objects.bulk_create_with_fallback(
            new_assets, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
        )
        logger.info("%s: Created %s child assets", character, len(created_objs))

    if new_assets and asset_data:
        # there are more child assets to create
        assets_create_children.apply_async(
            kwargs={
                "asset_list": list(asset_data.values()),
                "character_pk": character.pk,
                "cycle": cycle + 1,
            },
            priority=priority,
        )
    else:
        if len(asset_data) > 0:
            error_text = "child assets could not be added (orphans)"
            character.update_section_log_result(
                Character.UpdateSection.ASSETS,
                is_success=False,
                error_message=(f"{len(asset_data)} {error_text}"),
            )
            logger.warning("%s: %d %s", character, len(asset_data), error_text)

            # additional infos for analyzing issues #152
            logger.debug("Item IDs of orphans: %s", sorted(asset_data.keys()))
            orphan_location_ids = sorted(
                {item["location_id"] for item in asset_data.values()}
            )
            logger.debug("Location IDs of orphans: %s", orphan_location_ids)
            logger.debug("Parent asset items IDs: %s", sorted(parent_asset_ids))

        else:
            character.update_section_log_result(
                Character.UpdateSection.ASSETS, is_success=True, is_updated=True
            )


# Special tasks for updating mail section


@shared_task(**_update_task_params_with_bind)
def update_character_mails(self, character_pk: int, force_update: bool) -> None:
    """Update mails of a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    section = Character.UpdateSection.MAILS
    logger.info("%s: Updating %s", character, section.label)
    character.reset_update_section(section=section)
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    chain(
        update_character_mailing_lists.si(
            character.pk,
            force_update=force_update,
        ).set(priority=priority),
        update_character_mail_labels.si(
            character.pk,
            force_update=force_update,
        ).set(priority=priority),
        update_character_mails_headers_and_bodies.si(
            character.pk,
            force_update=force_update,
        ).set(priority=priority),
    ).delay()


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_character_mailing_lists(
    character_pk: int, force_update: bool = False
) -> None:
    """Update mailing list for a character."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    character.perform_update_with_error_logging(
        section=Character.UpdateSection.MAILS,
        method=character.update_mailing_lists,
        force_update=force_update,
    )


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_character_mail_labels(character_pk: int, force_update: bool = False) -> None:
    """Update mail labels for a character."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    character.perform_update_with_error_logging(
        section=Character.UpdateSection.MAILS,
        method=character.update_mail_labels,
        force_update=force_update,
    )


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
@when_esi_is_available
def update_character_mails_headers_and_bodies(
    self, character_pk: int, force_update: bool = False
) -> List[int]:
    """Update mail headers & bodies for a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    result = character.perform_update_with_error_logging(
        section=Character.UpdateSection.MAILS,
        method=character.update_mail_headers,
        force_update=force_update,
    )
    if result.is_changed or force_update:
        mail_ids = result.data.keys()
        mails_to_fetch = character.mails.filter(mail_id__in=mail_ids)
        if not force_update:
            mails_to_fetch = mails_to_fetch.filter(body="")

        priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
        logger.info(
            "%s: Loading %d mail bodies from ESI", character, mails_to_fetch.count()
        )
        for mail_id in mails_to_fetch.values_list("mail_id", flat=True):
            update_mail_body_esi.apply_async(
                kwargs={
                    "character_pk": character.pk,
                    "mail_id": mail_id,
                    "force_update": force_update,
                },
                priority=priority,
            )

    # the last task in the chain logs success (if any)
    character.update_section_log_result(
        Character.UpdateSection.MAILS, is_success=True, is_updated=result.is_updated
    )


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_mail_body_esi(character_pk: int, mail_id: int, force_update: bool = False):
    """Update the body of a character's mail from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    mail = character.mails.get(mail_id=mail_id)
    character.perform_update_with_error_logging(
        section=Character.UpdateSection.MAILS,
        method=character.update_mail_body,
        mail=mail,
        force_update=force_update,
    )


# special tasks for updating contacts


@shared_task(**_update_task_params_with_bind)
def update_character_contacts(self, character_pk: int, force_update: bool) -> None:
    """Update contacts of a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    section = Character.UpdateSection.CONTACTS
    character.reset_update_section(section=section)
    logger.info("%s: Updating %s", character, section.label)
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    chain(
        update_character_contact_labels.si(character.pk, force_update=force_update).set(
            priority=priority
        ),
        update_character_contacts_2.si(character.pk, force_update=force_update).set(
            priority=priority
        ),
    ).delay()


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_character_contact_labels(
    character_pk: int, force_update: bool = False
) -> None:
    """Update contact labels for a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    character.perform_update_with_error_logging(
        section=Character.UpdateSection.CONTACTS,
        method=character.update_contact_labels,
        force_update=force_update,
    )


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_character_contacts_2(character_pk: int, force_update: bool = False) -> None:
    """Update contacts for a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    result = character.perform_update_with_error_logging(
        section=Character.UpdateSection.CONTACTS,
        method=character.update_contacts,
        force_update=force_update,
    )
    character.update_section_log_result(
        Character.UpdateSection.CONTACTS, is_success=True, is_updated=result.is_updated
    )


# special tasks for updating contracts


@shared_task(**_update_task_params_with_bind)
def update_character_contracts(self, character_pk: int, force_update: bool) -> None:
    """Update contracts of a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    section = Character.UpdateSection.CONTRACTS
    character.reset_update_section(section=section)
    logger.info("%s: Updating %s", character, section.label)
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    chain(
        update_character_contract_headers.si(
            character.pk, force_update=force_update
        ).set(priority=priority),
        update_character_contracts_items.si(character.pk).set(priority=priority),
        update_character_contracts_bids.si(character.pk).set(priority=priority),
    ).delay()


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_character_contract_headers(character_pk: int, force_update: bool = False):
    """Update contract headers for a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    result = character.perform_update_with_error_logging(
        section=Character.UpdateSection.CONTRACTS,
        method=character.update_contract_headers,
        force_update=force_update,
    )
    character.update_section_log_result(
        Character.UpdateSection.CONTRACTS, is_success=True, is_updated=result.is_updated
    )


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
def update_character_contracts_items(self, character_pk: int):
    """Update items for all contracts of a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    contract_pks = set(
        character.contracts.filter(
            contract_type__in=[
                CharacterContract.TYPE_ITEM_EXCHANGE,
                CharacterContract.TYPE_AUCTION,
                # Courier contracts have no items
            ],
            items__isnull=True,
        ).values_list("pk", flat=True)
    )
    if len(contract_pks) > 0:
        logger.info(
            "%s: Starting updating items for %s contracts", character, len(contract_pks)
        )
        priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
        for contract_pk in contract_pks:
            update_contract_items_esi.apply_async(
                kwargs={"character_pk": character.pk, "contract_pk": contract_pk},
                priority=priority,
            )

    else:
        logger.info("%s: No items to update", character)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_contract_items_esi(character_pk: int, contract_pk: int):
    """Update the items of a character contract from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    contract = CharacterContract.objects.get(pk=contract_pk)
    character.update_contract_items(contract)


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
def update_character_contracts_bids(self, character_pk: int):
    """Update bids for all contracts of a character from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    contract_pks = set(
        character.contracts.filter(
            contract_type__in=[CharacterContract.TYPE_AUCTION],
            status=CharacterContract.STATUS_OUTSTANDING,
        ).values_list("pk", flat=True)
    )
    if len(contract_pks) > 0:
        logger.info(
            "%s: Starting updating bids for %s contracts", character, len(contract_pks)
        )
        priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
        for contract_pk in contract_pks:
            update_contract_bids_esi.apply_async(
                kwargs={"character_pk": character.pk, "contract_pk": contract_pk},
                priority=priority,
            )

    else:
        logger.info("%s: No bids to update", character)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_contract_bids_esi(character_pk: int, contract_pk: int):
    """Update bids of a character contract from ESI."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    contract = CharacterContract.objects.get(pk=contract_pk)
    character.update_contract_bids(contract)


# Tasks for other objects


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_market_prices():
    """Update market prices from ESI."""
    EveMarketPrice.objects.update_from_esi(
        minutes_until_stale=MEMBERAUDIT_SECTION_STALE_MINUTES_GLOBAL_DEFAULT
    )


@shared_task(
    **{
        **TASK_DEFAULTS_BIND_ONCE,
        **{
            "once": {"keys": ["id"], "graceful": True},
            "max_retries": MAX_RETRIES_STRUCTURES,
        },
    }
)
def update_structure_esi(self, id: int, token_pk: int):
    """Update a structure object from ESI.

    If the ESI error limit has already been reached retry later.
    """
    token = Token.objects.get(pk=token_pk)

    try:
        Location.objects.structure_update_or_create_esi(id, token)

    except EsiOffline as ex:
        backoff_jitter = int(random.uniform(2, 4) ** self.request.retries)
        countdown = (15 + backoff_jitter) * 60
        logger.warning(
            "Location #%s: ESI appears to be offline. Trying again in %d seconds.",
            id,
            countdown,
        )
        raise self.retry(countdown=countdown) from ex

    except EsiErrorLimitExceeded as ex:
        backoff_jitter = int(random.uniform(2, 4) ** self.request.retries)
        countdown = ex.retry_in + backoff_jitter
        logger.warning(
            "Location #%s: ESI error limit threshold reached. "
            "Trying again in %s seconds",
            id,
            countdown,
        )
        raise self.retry(countdown=countdown) from ex


@shared_task(
    **{
        **TASK_DEFAULTS_ONCE,
        **{
            "once": {"keys": ["id"], "graceful": True},
            "max_retries": MAX_RETRIES_MAIL_ENTITIES,
        },
    }
)
def update_mail_entity_esi(id: int, category: Optional[str] = None):
    """Update a mail entity object from ESI
    and retry later if the ESI error limit has already been reached.
    """
    MailEntity.objects.update_or_create_esi(id=id, category=category)


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
def update_characters_skill_checks(self, force_update: bool = False) -> None:
    """Start the update of skill checks for all registered characters.

    Args:
        - force_update: When set to True will always update regardless of stale status
    """
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    for character in Character.objects.all():
        if force_update or character.is_update_needed_for_section(
            Character.UpdateSection.SKILL_SETS
        ):
            update_character_skill_sets.apply_async(
                kwargs={"character_pk": character.pk, "force_update": force_update},
                priority=priority,
            )


@shared_task(**TASK_DEFAULTS_ONCE)
def check_character_consistency(character_pk) -> None:
    """Check consistency of a character."""
    character: Character = Character.objects.get_cached(
        pk=character_pk, timeout=MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT
    )
    character.update_sharing_consistency()


@shared_task(**TASK_DEFAULTS)
def delete_objects(model_name: str, obj_pks: Iterable[int]) -> None:
    """Delete multiple objects of a model."""
    model_class = apps.get_model("memberaudit", str(model_name))
    objs_to_delete = model_class.objects.filter(pk__in=obj_pks)
    amount = objs_to_delete.count()
    objs_to_delete.delete()
    logger.info("Deleted %d %s objects", amount, model_class.__name__)


@shared_task(**TASK_DEFAULTS_BIND)
def export_data(self, user_pk: Optional[int] = None) -> None:
    """Export data to files."""
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    tasks = [
        _export_data_for_topic.si(topic).set(priority=priority)
        for topic in data_exporters.DataExporter.topics()
    ]
    if user_pk:
        tasks.append(_export_data_inform_user.si(user_pk))
    chain(tasks).delay()


@shared_task(**TASK_DEFAULTS_BIND)
def export_data_for_topic(self, topic: str, user_pk: int):
    """Export data for a topic."""
    priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
    chain(
        _export_data_for_topic.si(topic).set(priority=priority),
        _export_data_inform_user.si(user_pk, topic).set(priority=priority),
    ).delay()


@shared_task(**TASK_DEFAULTS_ONCE)
def _export_data_for_topic(topic: str, destination_folder: Optional[str] = None) -> str:
    """Export data for given topic into a zipped file in destination."""
    file_path = data_exporters.export_topic_to_archive(
        topic=topic, destination_folder=destination_folder
    )
    return str(file_path)


@shared_task(**TASK_DEFAULTS)
def _export_data_inform_user(user_pk: int, topic: Optional[str] = None):
    user = User.objects.get(pk=user_pk)
    if topic:
        title = f"{__title__}: Data export for {topic} completed"
        message = f"Data export has been completed for topic {topic}."
    else:
        title = f"{__title__}: Full data export completed"
        message = (
            "Data export for all topics has been completed. "
            "It covers the following:\n"
        )
        for obj in data_exporters.DataExporter.topics():
            message += f"- {obj}\n"  # pylint: disable=consider-using-join
    notify(user=user, title=title, message=message, level="INFO")


@shared_task(**TASK_DEFAULTS_BIND_ONCE)
def update_compliance_groups_for_all(self):
    """Update compliance groups for all users."""
    if ComplianceGroupDesignation.objects.exists():
        priority = determine_task_priority(self) or MEMBERAUDIT_TASKS_LOW_PRIORITY
        for user in User.objects.all():
            update_compliance_groups_for_user.apply_async(
                kwargs={"user_pk": user.pk}, priority=priority
            )


@shared_task(**TASK_DEFAULTS)
def update_compliance_groups_for_user(user_pk: int):
    """Update compliance groups for user."""
    user = User.objects.get(pk=user_pk)
    ComplianceGroupDesignation.objects.update_user(user)


@shared_task(**TASK_DEFAULTS)
def add_compliant_users_to_group(group_pk: int):
    """Add compliant users to given group."""
    group = Group.objects.get(pk=group_pk)
    General.add_compliant_users_to_group(group)


@shared_task(**TASK_DEFAULTS)
def clear_users_from_group(group_pk: int):
    """Clear all users from given group."""
    group = Group.objects.get(pk=group_pk)
    utils.clear_users_from_group(group)



# Generated by Django 4.0.10 on 2023-09-28 12:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0011_add_standings_index"),
    ]

    operations = [
        migrations.AddField(
            model_name="character",
            name="token_error_notified_at",
            field=models.DateTimeField(
                default=None,
                help_text="Time when user was last notified about a token error.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="characterupdatestatus",
            name="has_token_error",
            field=models.BooleanField(
                default=False, help_text="Whether this section has a token error."
            ),
        ),
        migrations.AlterField(
            model_name="characterupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("assets", "assets"),
                    ("attributes", "attributes"),
                    ("character_details", "character details"),
                    ("contacts", "contacts"),
                    ("contracts", "contracts"),
                    ("corporation_history", "corporation history"),
                    ("fw_stats", "faction warfare statistics"),
                    ("implants", "implants"),
                    ("jump_clones", "jump clones"),
                    ("location", "location"),
                    ("loyalty", "loyalty"),
                    ("mails", "mails"),
                    ("mining_ledger", "mining ledger"),
                    ("online_status", "online status"),
                    ("planets", "planets"),
                    ("roles", "roles"),
                    ("ship", "ship"),
                    ("skills", "skills"),
                    ("skill_queue", "skill queue"),
                    ("skill_sets", "skill sets"),
                    ("standings", "standings"),
                    ("wallet_balance", "wallet balance"),
                    ("wallet_journal", "wallet journal"),
                    ("wallet_transactions", "wallet transactions"),
                ],
                db_index=True,
                max_length=64,
            ),
        ),
        migrations.CreateModel(
            name="CharacterRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "location",
                    models.CharField(
                        choices=[
                            ("BS", "base"),
                            ("HQ", "headquarters"),
                            ("OT", "other"),
                            ("UV", "universal"),
                        ],
                        help_text="Location where this role is applicable.",
                        max_length=2,
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[
                            ("AT1", "account take 1"),
                            ("AT2", "account take 2"),
                            ("AT3", "account take 3"),
                            ("AT4", "account take 4"),
                            ("AT5", "account take 5"),
                            ("AT6", "account take 6"),
                            ("AT7", "account take 7"),
                            ("ACT", "accountant"),
                            ("AUD", "auditor"),
                            ("COM", "communications officer"),
                            ("CEQ", "config equipment"),
                            ("CSE", "config starbase equipment"),
                            ("CT1", "container take 1"),
                            ("CT2", "container take 2"),
                            ("CT3", "container take 3"),
                            ("CT4", "container take 4"),
                            ("CT5", "container take 5"),
                            ("CT6", "container take 6"),
                            ("CT7", "container take 7"),
                            ("CMG", "contract manager"),
                            ("DPL", "diplomat"),
                            ("DRT", "director"),
                            ("FCM", "factory manager"),
                            ("FTM", "fitting manager"),
                            ("HQ1", "hangar query 1"),
                            ("HQ2", "hangar query 2"),
                            ("HQ3", "hangar query 3"),
                            ("HQ4", "hangar query 4"),
                            ("HQ5", "hangar query 5"),
                            ("HQ6", "hangar query 6"),
                            ("HQ7", "hangar query 7"),
                            ("HT1", "hangar take 1"),
                            ("HT2", "hangar take 2"),
                            ("HT3", "hangar take 3"),
                            ("HT4", "hangar take 4"),
                            ("HT5", "hangar take 5"),
                            ("HT6", "hangar take 6"),
                            ("HT7", "hangar take 7"),
                            ("JAC", "junior accountant"),
                            ("PSM", "personnel manager"),
                            ("RFF", "rent factory facility"),
                            ("RFC", "rent office"),
                            ("RRF", "rent research facility"),
                            ("SCO", "security officer"),
                            ("SPM", "skill plan manager"),
                            ("SDO", "starbase defense operator"),
                            ("SFT", "starbase fuel technician"),
                            ("STM", "station manager"),
                            ("TRD", "trader"),
                        ],
                        help_text="Role of a character",
                        max_length=3,
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="roles",
                        related_query_name="role",
                        to="memberaudit.character",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.AddConstraint(
            model_name="characterrole",
            constraint=models.UniqueConstraint(
                fields=("character", "location", "role"),
                name="functional_pk_characterrole",
            ),
        ),
    ]



"""Script for generating invalid locations to test fix for issue #153."""

from app_utils.scripts import start_django

start_django()


def main():
    from django.db.models import Max
    from eveuniverse.models import EveEntity, EveType

    from allianceauth.eveonline.models import EveCharacter

    from memberaudit.models import Character, Location
    from memberaudit.tests.testdata.constants import EveTypeId
    from memberaudit.tests.testdata.factories import (
        create_character,
        create_character_asset,
        create_character_contract_courier,
        create_character_jump_clone,
        create_character_jump_clone_implant,
        create_character_location,
        create_character_wallet_transaction,
        create_location,
    )

    EVE_CHARACTER_ID = 91224790  # CCP Seagull

    def max_location_id():
        max_location_id = Location.objects.aggregate(Max("id"))["id__max"]
        return max_location_id

    locations = Location.objects.filter(
        eve_type__isnull=False, eve_solar_system__isnull=False
    )
    valid_location_1 = locations.first()
    valid_location_2 = locations.last()
    if not valid_location_1 or valid_location_1 == valid_location_2:
        raise RuntimeError("Need two valid locations to exist")

    character_entity = EveEntity.objects.filter(
        category=EveEntity.CATEGORY_CHARACTER
    ).first()
    if not character_entity:
        raise RuntimeError("Need a character in eve entities to exist")

    corporation_entity = EveEntity.objects.filter(
        category=EveEntity.CATEGORY_CORPORATION
    ).first()
    if not corporation_entity:
        raise RuntimeError("Need a corporation in eve entities to exist")

    try:
        eve_character = EveCharacter.objects.get(character_id=EVE_CHARACTER_ID)
    except EveCharacter.DoesNotExist:
        eve_character = EveCharacter.objects.create_character(
            character_id=EVE_CHARACTER_ID
        )

    Character.objects.filter(eve_character=eve_character).delete()
    character = create_character(eve_character, is_disabled=True)

    item_eve_type, _ = EveType.objects.get_or_create_esi(id=EveTypeId.VELDSPAR)
    implant_eve_type, _ = EveType.objects.get_or_create_esi(
        id=EveTypeId.HIGH_GRADE_SNAKE_ALPHA
    )

    # create assets and invalid location
    asset_1 = create_character_asset(
        character=character,
        eve_type=item_eve_type,
        location=valid_location_1,
        item_id=max_location_id() + 1,
    )
    invalid_location_1 = create_location(id=asset_1.item_id)
    asset_2 = create_character_asset(
        character=character,
        eve_type=item_eve_type,
        location=valid_location_2,
        item_id=max_location_id() + 1,
    )
    invalid_location_2 = create_location(id=asset_2.item_id)

    # create character objects
    for location in [valid_location_1, invalid_location_1]:
        create_character_asset(
            character=character, eve_type=item_eve_type, location=location
        )
        jump_clone = create_character_jump_clone(character=character, location=location)
        create_character_jump_clone_implant(
            jump_clone=jump_clone, eve_type=implant_eve_type
        )
        create_character_wallet_transaction(
            character=character,
            location=location,
            eve_type=item_eve_type,
            client=corporation_entity,
        )

    create_character_location(character=character, location=invalid_location_1)
    create_character_contract_courier(
        character=character,
        assignee=None,
        issuer=character_entity,
        issuer_corporation=corporation_entity,
        start_location=valid_location_1,
        end_location=valid_location_2,
    )
    create_character_contract_courier(
        character=character,
        assignee=None,
        issuer=character_entity,
        issuer_corporation=corporation_entity,
        start_location=invalid_location_1,
        end_location=invalid_location_2,
    )
    for section in [
        Character.UpdateSection.ASSETS,
        Character.UpdateSection.CONTRACTS,
        Character.UpdateSection.LOCATION,
        Character.UpdateSection.JUMP_CLONES,
        Character.UpdateSection.WALLET_TRANSACTIONS,
    ]:
        character.update_section_log_result(
            section=section, is_success=True, is_updated=True
        )

    print("DONE!")


if __name__ == "__main__":
    main()



import json
from pathlib import Path
from unittest.mock import patch

import requests_mock

from app_utils.testing import NoSocketsTestCase

from memberaudit.core import esi_status
from memberaudit.models import Character

MODULE_PATH = "memberaudit.core.esi_status"


class TestEndpoint(NoSocketsTestCase):
    def test_should_not_allow_invalid_creation(self):
        cases = [
            ("", ""),
            ("xxx", "/characters/{character_id}"),
            ("", "/characters/{character_id}"),
            ("get", ""),
        ]
        for method, route in cases:
            with self.subTest(method=method, route=route):
                with self.assertRaises(ValueError):
                    esi_status._Endpoint(method=method, route=route)

    def test_can_create_from_dict(self):
        ep = esi_status._Endpoint.from_dict(
            {
                "endpoint": "esi-assets",
                "method": "get",
                "route": "/characters/{character_id}/assets/",
                "status": "green",
                "tags": ["Assets"],
            },
        )
        self.assertEqual(ep.method, "get")
        self.assertEqual(ep.route, "/characters/{character_id}/assets/")


@patch(MODULE_PATH + "._unavailable_sections", spec=True)
@patch(MODULE_PATH + ".cache.set", spec=True)
@patch(MODULE_PATH + ".cache.get", spec=True)
class TestUnavailableSections(NoSocketsTestCase):
    def test_should_return_from_cache(
        self, mock_cache_get, mock_cache_set, mock_unavailable_sections
    ):
        mock_cache_get.return_value = {Character.UpdateSection.ASSETS}
        x = esi_status.unavailable_sections()
        self.assertSetEqual(x, {Character.UpdateSection.ASSETS})

    def test_should_update_cache_and_return_new_value(
        self, mock_cache_get, mock_cache_set, mock_unavailable_sections
    ):
        mock_cache_get.return_value = None
        mock_unavailable_sections.return_value = {Character.UpdateSection.ASSETS}
        x = esi_status.unavailable_sections()
        self.assertSetEqual(x, {Character.UpdateSection.ASSETS})
        self.assertTrue(mock_cache_set.called)

    def test_should_none_on_failure(
        self, mock_cache_get, mock_cache_set, mock_unavailable_sections
    ):
        mock_cache_get.return_value = None
        mock_unavailable_sections.return_value = None
        x = esi_status.unavailable_sections()
        self.assertIsNone(x)
        self.assertFalse(mock_cache_set.called)


@requests_mock.Mocker()
class TestUnavailableSections2(NoSocketsTestCase):
    def test_should_return_unavailable_sections_as_reported_by_ESI(
        self, requests_mocker
    ):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            json=[
                {
                    "endpoint": "esi-mail",
                    "method": "get",
                    "route": "/characters/{character_id}/mail/",
                    "status": "green",
                    "tags": ["Mail"],
                },
                {
                    "endpoint": "esi-loyalty",
                    "method": "get",
                    "route": "/characters/{character_id}/loyalty/points/",
                    "status": "red",
                    "tags": ["Loyalty"],
                },
                {
                    "endpoint": "esi-loyalty",
                    "method": "get",
                    "route": "/characters/{character_id}/loyalty/points/xy/",
                    "status": "green",
                    "tags": ["Loyalty"],
                },
            ],
        )
        # when
        got = esi_status._unavailable_sections()
        # then
        want = {Character.UpdateSection.LOYALTY}
        self.assertEqual(want, got)

    def test_should_return_an_empty_set_when_all_sections_available(
        self, requests_mocker
    ):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            json=[
                {
                    "endpoint": "esi-mail",
                    "method": "get",
                    "route": "/characters/{character_id}/mail/",
                    "status": "green",
                    "tags": ["Mail"],
                },
                {
                    "endpoint": "esi-loyalty",
                    "method": "get",
                    "route": "/characters/{character_id}/loyalty/points/",
                    "status": "green",
                    "tags": ["Loyalty"],
                },
            ],
        )
        # when
        got = esi_status._unavailable_sections()
        # then
        want = set()
        self.assertEqual(want, got)

    def test_should_report_when_esi_status_could_not_be_fetched(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            status_code=500,
        )
        # when
        got = esi_status._unavailable_sections()
        # then
        self.assertIsNone(got)

    def test_should_return_as_error_when_no_endpoints_are_returned(
        self, requests_mocker
    ):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            json=[],
        )
        # when
        got = esi_status._unavailable_sections()
        # then
        self.assertIsNone(got)


@requests_mock.Mocker()
class TestFetchStatus(NoSocketsTestCase):
    def test_can_fetch_status(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            json=[
                {
                    "endpoint": "esi-mail",
                    "method": "get",
                    "route": "/characters/{character_id}/mail/",
                    "status": "green",
                    "tags": ["Mail"],
                }
            ],
        )
        # when
        got = esi_status._fetch_status()
        # then
        want = [
            {
                "endpoint": "esi-mail",
                "method": "get",
                "route": "/characters/{character_id}/mail/",
                "status": "green",
                "tags": ["Mail"],
            }
        ]
        self.assertListEqual(want, got)

    def test_should_report_http_error(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            status_code=500,
        )
        # when
        got = esi_status._fetch_status()
        # then
        self.assertIsNone(got)

    def test_should_report_json_error(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            text="this is not json",
        )
        # when
        got = esi_status._fetch_status()
        # then
        self.assertIsNone(got)


@patch(MODULE_PATH + ".sleep", lambda x: None)
@requests_mock.Mocker()
class TestGetEsiStatus(NoSocketsTestCase):
    def test_should_return_response_when_ok(self, requests_mocker):
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            text="ok",
        )
        # when
        got = esi_status._get_esi_status()
        # then
        self.assertTrue(got.ok)
        self.assertEqual(got.text, "ok")
        self.assertEqual(requests_mocker.call_count, 1)

    def test_should_return_most_errors_directly(self, requests_mocker):
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            status_code=500,
        )
        # when
        got = esi_status._get_esi_status()
        # then
        self.assertEqual(got.status_code, 500)
        self.assertEqual(requests_mocker.call_count, 1)

    def test_should_retry_on_specific_errors(self, requests_mocker):
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            status_code=503,
        )
        # when
        got = esi_status._get_esi_status()
        # then
        self.assertEqual(got.status_code, 503)
        self.assertEqual(requests_mocker.call_count, 3)


class TestSectionEndpointsDef(NoSocketsTestCase):
    def test_all_sections_must_have_endpoints_defined(self):
        excluded = {Character.UpdateSection.SKILL_SETS}
        for s in Character.UpdateSection:
            if s in excluded:
                continue
            if s not in esi_status._REQUIRED_ENDPOINTS_FOR_SECTIONS:
                self.fail(f"does not cover section: {s}")
            if len(esi_status._REQUIRED_ENDPOINTS_FOR_SECTIONS[s]) == 0:
                self.fail(f"missing endpoints definition for section: {s}")

    def test_section_endpoints_must_be_valid(self):
        # given
        p = Path(__file__).parent / "esi_status_example.json"
        with p.open("r", encoding="utf8") as f:
            status = json.load(f)
        valid_endpoints = {(ep["method"], ep["route"]) for ep in status}
        for s, endpoints in esi_status._REQUIRED_ENDPOINTS_FOR_SECTIONS.items():
            endpoints: list[esi_status._Endpoint]
            for ep in endpoints:
                if (ep.method, ep.route) not in valid_endpoints:
                    self.fail(f"{s}: invalid route: {ep}")



from unittest.mock import patch

from django.test import override_settings, tag
from django.urls import reverse
from django_webtest import WebTest
from eveuniverse.models import EveType

from allianceauth.tests.auth_utils import AuthUtils

from memberaudit.models import Character, CharacterContract, Location, MailEntity
from memberaudit.tests.testdata.esi_client_stub import esi_stub
from memberaudit.tests.testdata.factories import (
    create_character_asset,
    create_character_contract,
    create_character_contract_item,
    create_character_mail,
    create_character_mail_label,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import (
    add_auth_character_to_user,
    add_memberaudit_character_to_user,
    create_memberaudit_character,
    create_user_from_evecharacter_with_access,
    reset_celery_once_locks,
)

MANAGERS_PATH = "memberaudit.managers"
MODELS_PATH = "memberaudit.models"
TASKS_PATH = "memberaudit.tasks"


class TestUILauncher(WebTest):
    fixtures = ["disable_analytics.json"]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        reset_celery_once_locks()

    def setUp(self) -> None:
        self.user, _ = create_user_from_evecharacter_with_access(1002)

    def test_open_character_viewer(self):
        """
        given user has character registered
        when clicking on respective character link
        then user is forwarded to character viewer
        """
        # setup
        character = add_memberaudit_character_to_user(self.user, 1001)

        # login & open launcher page
        self.app.set_user(self.user)
        launcher = self.app.get(reverse("memberaudit:launcher"))
        self.assertEqual(launcher.status_code, 200)

        # user clicks on character link
        character_viewer = launcher.click(
            href=reverse("memberaudit:character_viewer", args=[character.pk]),
            index=0,  # follow the first matching link
        )
        self.assertEqual(character_viewer.status_code, 200)

    @tag("breaks_with_older_mariadb")  # FIXME
    @patch(TASKS_PATH + ".esi_status.unavailable_sections", lambda: set())
    @patch(MANAGERS_PATH + ".character_sections_1.esi", esi_stub)
    @patch(MANAGERS_PATH + ".character_sections_2.esi", esi_stub)
    @patch(MANAGERS_PATH + ".character_sections_3.esi", esi_stub)
    @patch(MANAGERS_PATH + ".general.esi", esi_stub)
    @override_settings(
        CELERY_ALWAYS_EAGER=True,
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        APP_UTILS_OBJECT_CACHE_DISABLED=True,
    )
    def test_add_character(self):
        """
        when clicking on "register"
        then user can add a new character
        """
        # user as another auth character
        character_ownership_1001 = add_auth_character_to_user(self.user, 1001)

        # login & open launcher page
        self.app.set_user(self.user)
        launcher = self.app.get(reverse("memberaudit:launcher"))
        self.assertEqual(launcher.status_code, 200)

        # user clicks on register link
        select_token = launcher.click(
            href=reverse("memberaudit:add_character"),
            index=1,  # follow the 2nd matching link
        )
        self.assertEqual(select_token.status_code, 200)

        # user selects auth character 1001
        token = self.user.token_set.get(character_id=1001)
        my_form = None
        for form in select_token.forms.values():
            try:
                if int(form["_token"].value) == token.pk:
                    my_form = form
                    break
            except AssertionError:
                pass

        self.assertIsNotNone(my_form)
        launcher = my_form.submit().follow()
        self.assertEqual(launcher.status_code, 200)

        # check update went through
        character_1001: Character = (
            character_ownership_1001.character.memberaudit_character
        )
        self.assertTrue(character_1001.is_update_status_ok())

        # check added character is now visible in launcher
        a_tags = launcher.html.find_all("a", href=True)
        viewer_url = reverse("memberaudit:character_viewer", args=[character_1001.pk])
        character_1001_links = [
            a_tag["href"] for a_tag in a_tags if a_tag["href"] == viewer_url
        ]
        self.assertGreater(len(character_1001_links), 0)

    def test_share_character_1(self):
        """
        when user has share permission
        then he can share his characters
        """
        # setup
        character_1001 = add_memberaudit_character_to_user(self.user, 1001)
        self.user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters", self.user
        )

        # login & open launcher page
        self.app.set_user(self.user)
        launcher = self.app.get(reverse("memberaudit:launcher"))
        self.assertEqual(launcher.status_code, 200)

        # check for share button
        share_url = reverse("memberaudit:share_character", args=[character_1001.pk])
        a_tags = launcher.html.find_all("a", href=True)
        character_1001_links = [
            a_tag["href"] for a_tag in a_tags if a_tag["href"] == share_url
        ]
        self.assertGreater(len(character_1001_links), 0)

    def test_share_character_2(self):
        """
        when user does not have share permission
        then he can not share his characters
        """
        # setup
        character_1001 = add_memberaudit_character_to_user(self.user, 1001)

        # login & open launcher page
        self.app.set_user(self.user)
        launcher = self.app.get(reverse("memberaudit:launcher"))
        self.assertEqual(launcher.status_code, 200)

        # check for share button
        share_url = reverse("memberaudit:share_character", args=[character_1001.pk])
        a_tags = launcher.html.find_all("a", href=True)
        character_1001_links = [
            a_tag["href"] for a_tag in a_tags if a_tag["href"] == share_url
        ]
        self.assertEqual(len(character_1001_links), 0)


class TestUICharacterViewer(WebTest):
    fixtures = ["disable_analytics.json"]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character = create_memberaudit_character(1001)
        cls.user = cls.character.eve_character.character_ownership.user
        cls.jita_44 = Location.objects.get(id=60003760)

    def test_asset_container(self):
        """
        given user has a registered character with assets which contain other assets
        when user clicks on an asset container
        then the contents of that asset container are shown
        """
        # setup data
        parent_asset = create_character_asset(
            character=self.character,
            location=self.jita_44,
            eve_type=EveType.objects.get(name="Charon"),
        )
        create_character_asset(
            character=self.character,
            parent=parent_asset,
            eve_type=EveType.objects.get(name="Merlin"),
        )

        # open character viewer
        self.app.set_user(self.user)
        character_viewer = self.app.get(
            reverse("memberaudit:character_viewer", args=[self.character.pk])
        )
        self.assertEqual(character_viewer.status_code, 200)

        # open asset container
        asset_container = self.app.get(
            reverse(
                "memberaudit:character_asset_container",
                args=[self.character.pk, parent_asset.pk],
            )
        )
        self.assertEqual(asset_container.status_code, 200)
        self.assertIn("Asset Container", asset_container.text)

    def test_contract_items(self):
        """
        given user has a registered character with contracts that contain items
        when user clicks to open the contract
        then the items of that contact are shown
        """
        # setup data
        contract = create_character_contract(
            character=self.character,
            contract_type=CharacterContract.TYPE_ITEM_EXCHANGE,
            start_location=self.jita_44,
            end_location=self.jita_44,
        )
        create_character_contract_item(
            contract=contract,
            eve_type=EveType.objects.get(name="High-grade Snake Alpha"),
        )

        # open character viewer
        self.app.set_user(self.user)
        character_viewer = self.app.get(
            reverse("memberaudit:character_viewer", args=[self.character.pk])
        )
        self.assertEqual(character_viewer.status_code, 200)

        # open asset container
        contract_details = self.app.get(
            reverse(
                "memberaudit:character_contract_details",
                args=[self.character.pk, contract.pk],
            )
        )
        self.assertEqual(contract_details.status_code, 200)
        self.assertIn("High-grade Snake Alpha", contract_details.text)

    def test_mail(self):
        """
        given user has a registered character with mails
        when user clicks to open a mail
        then the mail body is shown
        """
        # setup data
        body_text = "My text body"
        label = create_character_mail_label(character=self.character)
        sender_1002, _ = MailEntity.objects.update_or_create_from_eve_entity_id(id=1002)
        mail = create_character_mail(
            character=self.character, sender=sender_1002, body=body_text
        )
        recipient_1001, _ = MailEntity.objects.update_or_create_from_eve_entity_id(
            id=1001
        )
        recipient_1003, _ = MailEntity.objects.update_or_create_from_eve_entity_id(
            id=1003
        )
        mail.recipients.add(recipient_1001, recipient_1003)
        mail.labels.add(label)

        # open character viewer
        self.app.set_user(self.user)
        character_viewer = self.app.get(
            reverse("memberaudit:character_viewer", args=[self.character.pk])
        )
        self.assertEqual(character_viewer.status_code, 200)

        # open mail
        mail_details = self.app.get(
            reverse("memberaudit:character_mail", args=[self.character.pk, mail.pk])
        )
        self.assertEqual(mail_details.status_code, 200)
        self.assertIn(body_text, mail_details.text)



# Generated by Django 4.0.10 on 2024-01-17 22:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0018_charactercloneinfo"),
    ]

    operations = [
        migrations.AlterField(
            model_name="charactercontractitem",
            name="record_id",
            field=models.PositiveBigIntegerField(db_index=True),
        ),
    ]



"""Parser for fitting in EFT Format"""

from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from bravado.exception import HTTPNotFound

from eveuniverse.models import EveEntity, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.constants import EveCategoryId, EveDogmaEffectId, EveGroupId

from .fittings import Fitting, Item, Module

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EftParserError(Exception):
    """Base error for EFT parser."""


class UnknownShipTypeError(EftParserError):
    """Ship type is unknown."""


class MissingTitleError(EftParserError):
    """Title is missing."""


class MissingSectionsError(EftParserError):
    """Insufficient sections defined."""


@dataclass
class _EveTypes:
    """Container with EveType objects to enable quick name to object resolution."""

    objs_by_name: Dict[str, EveType] = field(default_factory=dict)

    def from_name(self, type_name: Optional[str]) -> Optional[EveType]:
        """Resolve given type name into EveType object.

        Returns ``None`` if it can not be resolved.
        """
        if type_name is None:
            return None
        return self.objs_by_name.get(str(type_name))

    @classmethod
    def create_from_names(
        cls, type_names: Iterable[str]
    ) -> Tuple["_EveTypes", Set[str]]:
        """Create new object from list of type names.

        Will try to fetch types from DB first and load missing types from ESI.
        All types must have dogmas.

        Returns:
            created object, list of type names that could not be resolved
        """
        if not isinstance(type_names, (set, list, tuple)):
            raise TypeError("type_name must be of type list, set or tuple.")
        type_names = set(type_names)
        eve_types = cls._fetch_types_from_db(type_names)
        missing_type_names = type_names - set(eve_types.keys())
        if missing_type_names:
            added_eve_types, missing_type_names_2 = cls._fetch_missing_types_from_esi(
                missing_type_names
            )
            eve_types.update(added_eve_types)
        else:
            missing_type_names_2 = set()
        return cls(eve_types), missing_type_names_2

    @classmethod
    def _fetch_types_from_db(cls, type_names: Iterable[str]) -> Dict[str, EveType]:
        eve_types_query = (
            EveType.objects.select_related("eve_group")
            .prefetch_related("dogma_effects")
            .filter(
                enabled_sections=EveType.enabled_sections.dogmas, name__in=type_names  # type: ignore
            )
        )
        eve_types = {obj.name: obj for obj in eve_types_query}
        return eve_types

    @classmethod
    def _fetch_missing_types_from_esi(
        cls,
        missing_type_names: Set[str],
    ) -> Tuple[Dict[str, EveType], Set[str]]:
        def type_names_str(type_names: Iterable) -> str:
            return ", ".join(sorted(list(type_names)))

        logger.info(
            "EFT parser: trying to fetch unknown types from ESI: %s",
            type_names_str(missing_type_names),
        )
        entity_ids = (
            EveEntity.objects.fetch_by_names_esi(missing_type_names)
            .filter(category=EveEntity.CATEGORY_INVENTORY_TYPE)
            .values_list("id", flat=True)
        )
        eve_types = cls._fetch_types_from_esi(entity_ids)
        missing_type_names_2 = missing_type_names - set(eve_types.keys())
        if missing_type_names_2:
            logger.info(
                "EFT parser: failed to identify types: %s",
                type_names_str(missing_type_names_2),
            )
        return eve_types, missing_type_names_2

    @staticmethod
    def _fetch_types_from_esi(entity_ids) -> Dict[str, EveType]:
        eve_types = {}
        for entity_id in entity_ids:
            try:
                obj, _ = EveType.objects.get_or_create_esi(
                    id=entity_id, enabled_sections=[EveType.Section.DOGMAS]
                )
            except HTTPNotFound:
                pass
            else:
                eve_types[obj.name] = obj  # type: ignore
        return eve_types

    # @classmethod
    # def _fetch_types_from_esi(cls, entity_ids) -> Dict[str, EveType]:
    #     """Fetch types from ESI concurrently using threads."""
    #     max_workers = getattr(settings, "ESI_CONNECTION_POOL_MAXSIZE", 10)
    #     with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
    #         eve_types = executor.map(cls._fetch_type_from_esi, entity_ids)
    #     return {obj.name: obj for obj in eve_types if obj}

    # @staticmethod
    # def _fetch_type_from_esi(entity_id) -> Optional[EveType]:
    #     """Fetch type from ESI."""
    #     try:
    #         obj, _ = EveType.objects.get_or_create_esi(
    #             id=entity_id, enabled_sections=[EveType.Section.DOGMAS]
    #         )
    #     except HTTPNotFound:
    #         return None
    #     else:
    #         return obj


class _EftSlotType(Enum):
    """An EFT slot type."""

    NONE = auto()
    LOW_SLOT = auto()
    MEDIUM_SLOT = auto()
    HIGH_SLOT = auto()
    RIG_SLOT = auto()
    SUBSYSTEM_SLOT = auto()


@dataclass
class _EftTextItem:
    """Text item of an EFT fitting used for parsing."""

    item_type: Optional[str] = None
    charge_type: Optional[str] = None
    quantity: Optional[int] = None
    is_offline: bool = False
    slot_type: _EftSlotType = _EftSlotType.NONE

    def type_names(self) -> Set[str]:
        """Return types used in the item."""
        types = set()
        if self.item_type:
            types.add(self.item_type)
        if self.charge_type:
            types.add(self.charge_type)
        return types

    @classmethod
    def create_from_line(cls, line: str) -> "_EftTextItem":
        """Create new object from text line."""
        empty_line_parsed = line.strip("[]").lower()
        if "empty " in empty_line_parsed:
            if " high " in empty_line_parsed:
                return cls(slot_type=_EftSlotType.HIGH_SLOT)
            if " med " in empty_line_parsed:
                return cls(slot_type=_EftSlotType.MEDIUM_SLOT)
            if " low " in empty_line_parsed:
                return cls(slot_type=_EftSlotType.LOW_SLOT)
            if " rig " in empty_line_parsed:
                return cls(slot_type=_EftSlotType.RIG_SLOT)
            if " subsystem " in empty_line_parsed:
                return cls(slot_type=_EftSlotType.SUBSYSTEM_SLOT)
            return cls()
        part = line.split()[-1]
        if "x" in part and part[1:].isdigit():
            item_type = line.split(part)[0].strip()
            quantity = part[1:]
            return cls(item_type=item_type, quantity=int(quantity))
        if "/OFFLINE" in line:
            is_offline = True
            line = line.replace(" /OFFLINE", "")
        else:
            is_offline = False
        if "," in line:
            item_type, charge_type = line.split(",")
            charge_type = charge_type.strip()
            return cls(
                item_type=item_type, charge_type=charge_type, is_offline=is_offline
            )
        return cls(item_type=line.strip(), is_offline=is_offline)


@dataclass
class _EftTextSection:
    """Text section of an EFT fitting used for parsing."""

    items: List[_EftTextItem] = field(default_factory=list)

    def type_names(self) -> Set[str]:
        """Return types used in the section."""
        types = set()
        for item in self.items:
            types |= item.type_names()
        return types

    @classmethod
    def create_from_lines(cls, lines: Iterable[str]) -> "_EftTextSection":
        """Create new object from lines."""
        items = [_EftTextItem.create_from_line(line) for line in lines]
        return cls(items=items)


@dataclass
class _EftItem:
    """Item of an EFT fitting used for parsing."""

    item_type: Optional[EveType] = None
    charge_type: Optional[EveType] = None
    quantity: Optional[int] = None
    is_offline: bool = False
    slot_type: _EftSlotType = _EftSlotType.NONE

    @property
    def is_empty(self) -> bool:
        """Return True if this item is empty, else False."""
        return self.item_type is None

    @property
    def is_slot(self) -> bool:
        """Return True if this item is a slot, else False."""
        return self.quantity is None

    def is_booster(self) -> bool:
        """Return True if this item is a booster, else False."""
        if self.is_empty:
            return False
        return self._item_group_id() == EveGroupId.BOOSTER

    def is_cyber_implant(self) -> bool:
        """Return True if this item is a cyber slot, else False."""
        if self.is_empty:
            return False
        return self._item_group_id() == EveGroupId.CYBERIMPLANT

    def is_drone(self) -> bool:
        """Return True if this item is a drone, else False."""
        if self.is_empty:
            return False
        return self._item_category_id() == EveCategoryId.DRONE

    def is_fighter(self) -> bool:
        """Return True if this item is a fighter, else False."""
        if self.is_empty:
            return False
        return self._item_category_id() == EveCategoryId.FIGHTER

    def is_high_slot(self) -> bool:
        """Return True if this item is a high slot, else False."""
        if self.slot_type is _EftSlotType.HIGH_SLOT:
            return True

        if self.is_empty:
            return False

        effect_ids = {
            obj.eve_dogma_effect_id for obj in self.item_type.dogma_effects.all()
        }
        return EveDogmaEffectId.HI_POWER in effect_ids

    def is_med_slot(self) -> bool:
        """Return True if this item is a med slot, else False."""
        if self.slot_type is _EftSlotType.MEDIUM_SLOT:
            return True

        if self.is_empty:
            return False

        effect_ids = {
            obj.eve_dogma_effect_id for obj in self.item_type.dogma_effects.all()
        }
        return EveDogmaEffectId.MED_POWER in effect_ids

    def is_low_slot(self) -> bool:
        """Return True if this item is a low slot, else False."""
        if self.slot_type is _EftSlotType.LOW_SLOT:
            return True

        if self.is_empty:
            return False

        effect_ids = {
            obj.eve_dogma_effect_id for obj in self.item_type.dogma_effects.all()
        }
        return EveDogmaEffectId.LO_POWER in effect_ids

    def is_rig_slot(self) -> bool:
        """Return True if this item is a rig slot."""
        if self.slot_type is _EftSlotType.RIG_SLOT:
            return True

        if self.is_empty:
            return False

        effect_ids = {
            obj.eve_dogma_effect_id for obj in self.item_type.dogma_effects.all()
        }
        return EveDogmaEffectId.RIG_SLOT in effect_ids

    def is_subsystem(self) -> bool:
        """Return True if this item is a subsystem, else False."""
        if self.slot_type is _EftSlotType.SUBSYSTEM_SLOT:
            return True

        if self.is_empty:
            return False

        effect_ids = {
            obj.eve_dogma_effect_id for obj in self.item_type.dogma_effects.all()
        }
        return EveDogmaEffectId.SUB_SYSTEM in effect_ids

    def _item_category_id(self) -> Optional[int]:
        if self.is_empty:
            return None
        return self.item_type.eve_group.eve_category_id

    def _item_group_id(self) -> Optional[int]:
        if self.is_empty:
            return None
        return self.item_type.eve_group_id

    @classmethod
    def create_from_text_item(
        cls, text_item: _EftTextItem, eve_types: _EveTypes
    ) -> "_EftItem":
        """Create new item from text."""
        return cls(
            item_type=eve_types.from_name(text_item.item_type),
            charge_type=eve_types.from_name(text_item.charge_type),
            quantity=text_item.quantity,
            is_offline=text_item.is_offline,
            slot_type=text_item.slot_type,
        )


@dataclass
class _EftSection:
    """Section of an EFT fitting used for parsing."""

    class Category(Enum):
        """A category in an EFT section."""

        UNKNOWN = auto()
        HIGH_SLOTS = auto()
        MEDIUM_SLOTS = auto()
        LOW_SLOTS = auto()
        RIG_SLOTS = auto()
        SUBSYSTEM_SLOTS = auto()
        DRONES_BAY = auto()
        FIGHTER_BAY = auto()
        IMPLANTS = auto()
        BOOSTERS = auto()
        CARGO_BAY = auto()

        @property
        def is_slots(self) -> bool:
            """Return True, if this category has slots."""
            return self in {
                self.HIGH_SLOTS,
                self.MEDIUM_SLOTS,
                self.LOW_SLOTS,
                self.RIG_SLOTS,
            }

    items: List[_EftItem] = field(default_factory=list)
    category: Category = Category.UNKNOWN

    @property
    def is_slots(self) -> bool:
        """Return True if this any item has slots."""
        return any((item.is_slot for item in self.items))

    def guess_category(self) -> Optional["_EftSection.Category"]:
        """Try to guess the category of this section based on it's items.
        Returns ``None`` if the guess fails.
        """
        if self.is_slots:
            if any((item.is_booster() for item in self.items)):
                return self.Category.BOOSTERS
            if any((item.is_cyber_implant() for item in self.items)):
                return self.Category.IMPLANTS
            if any((item.is_low_slot() for item in self.items)):
                return self.Category.LOW_SLOTS
            if any((item.is_med_slot() for item in self.items)):
                return self.Category.MEDIUM_SLOTS
            if any((item.is_high_slot() for item in self.items)):
                return self.Category.HIGH_SLOTS
            if any((item.is_rig_slot() for item in self.items)):
                return self.Category.RIG_SLOTS
            if any((item.is_subsystem() for item in self.items)):
                return self.Category.SUBSYSTEM_SLOTS
        else:
            if any((item.is_drone() for item in self.items)):
                return self.Category.DRONES_BAY
            if any((item.is_fighter() for item in self.items)):
                return self.Category.FIGHTER_BAY
        return None

    def to_modules(self) -> List[Module]:
        """Convert eft items into fitting modules.

        Types from modules that can not be resolved will result in an empty slot.
        """
        objs = []
        for item in self.items:
            if item.is_empty:
                objs.append(Module())
            else:
                params = {"module_type": item.item_type, "is_offline": item.is_offline}
                if item.charge_type:
                    params["charge_type"] = item.charge_type
                objs.append(Module(**params))
        return objs

    def to_items(self) -> List[Item]:
        """Convert eft items into fitting items."""
        objs = []
        for item in self.items:
            params: Dict[str, Any] = {"item_type": item.item_type}
            if item.quantity:
                params["quantity"] = item.quantity
            objs.append(Item(**params))
        return objs

    @classmethod
    def create_from_eft_text_section(
        cls, text_section: _EftTextSection, eve_types: _EveTypes
    ) -> "_EftSection":
        """Create new object from an EFT text section."""
        items = [
            _EftItem.create_from_text_item(item, eve_types)
            for item in text_section.items
        ]
        return cls(items=items)


def create_fitting_from_eft(eft_text: str) -> Tuple[Fitting, List[str]]:
    """Create new object from fitting in EFT format."""
    lines = _text_to_lines(eft_text)
    text_sections = _lines_to_text_sections(lines)
    eft_sections = [
        _EftTextSection.create_from_lines(lines=lines) for lines in text_sections
    ]
    ship_type_name, fitting_name = _parse_title(lines)
    eve_types, unknown_types = _load_eve_types(ship_type_name, eft_sections)
    sections = [
        _EftSection.create_from_eft_text_section(section, eve_types)
        for section in eft_sections
    ]
    sections = _try_to_identify_sections(sections)
    ship_type = eve_types.from_name(ship_type_name)
    if not ship_type:
        raise UnknownShipTypeError(f"Unknown ship type: {ship_type_name}")
    fitting = _create_fitting_from_sections(fitting_name, ship_type, sections)
    errors = _unknown_types_to_errors(unknown_types)
    return fitting, errors


def _text_to_lines(eft_text: str) -> List[str]:
    """Convert text into lines."""
    lines = eft_text.strip().splitlines()
    if not lines:
        raise MissingSectionsError("Text is empty")
    return lines


def _lines_to_text_sections(lines: List[str]) -> List[List[str]]:
    """Split lines into text sections."""
    text_sections = []
    section_lines = []
    for line in lines[1:]:
        if line:
            section_lines.append(line)
        else:
            if section_lines:
                text_sections.append(section_lines)
                section_lines = []
    if section_lines:
        text_sections.append(section_lines)
    return text_sections


def _parse_title(lines: List[str]) -> Tuple[str, str]:
    """Try to parse title from lines."""
    if not lines:
        raise MissingSectionsError("Text is empty")
    line = lines[0]
    if line.startswith("[") and "," in line:
        ship_type_name, fitting_name = line[1:-1].split(",")
        return ship_type_name.strip(), fitting_name.strip()
    raise MissingTitleError("Title not found")


def _load_eve_types(
    ship_type_name: str, sections: List[_EftTextSection]
) -> Tuple["_EveTypes", Set[str]]:
    """Load all EveType objects used in this fitting."""
    type_names = {ship_type_name}
    for section in sections:
        type_names |= section.type_names()
    return _EveTypes.create_from_names(type_names)


def _try_to_identify_sections(sections: List[_EftSection]) -> List[_EftSection]:
    """Identify sections if possible."""
    for section in sections:
        category = section.guess_category()
        if category:
            section.category = category
    # last unknown section must be the cargo bay
    if sections:
        last_section = sections[len(sections) - 1]
        if (
            last_section.category == _EftSection.Category.UNKNOWN
            and not last_section.is_slots
        ):
            last_section.category = _EftSection.Category.CARGO_BAY
    return sections


def _create_fitting_from_sections(
    fitting_name: str, ship_type: EveType, sections: List[_EftSection]
) -> Fitting:
    """Create fitting object from input."""
    params = {"name": fitting_name, "ship_type": ship_type}
    for section in sections:
        if section.category == _EftSection.Category.HIGH_SLOTS:
            params["high_slots"] = section.to_modules()
        elif section.category == _EftSection.Category.MEDIUM_SLOTS:
            params["medium_slots"] = section.to_modules()
        elif section.category == _EftSection.Category.LOW_SLOTS:
            params["low_slots"] = section.to_modules()
        elif section.category == _EftSection.Category.RIG_SLOTS:
            params["rig_slots"] = section.to_modules()
        elif section.category == _EftSection.Category.SUBSYSTEM_SLOTS:
            params["subsystem_slots"] = section.to_modules()
        elif section.category == _EftSection.Category.DRONES_BAY:
            params["drone_bay"] = section.to_items()
        elif section.category == _EftSection.Category.FIGHTER_BAY:
            params["fighter_bay"] = section.to_items()
        elif section.category == _EftSection.Category.IMPLANTS:
            params["implants"] = section.to_items()
        elif section.category == _EftSection.Category.BOOSTERS:
            params["boosters"] = section.to_items()
        elif section.category == _EftSection.Category.CARGO_BAY:
            params["cargo_bay"] = section.to_items()
    return Fitting(**params)


def _unknown_types_to_errors(unknown_types: Set[str]) -> List[str]:
    errors = []
    if unknown_types:
        names = ", ".join(sorted(list(unknown_types)))
        errors.append(f"Ignored modules/items with unknown types: {names}")
    return errors



"""Helpers for commands."""

from enum import Enum
from typing import Any, Dict, List, Tuple, Union

from django.core.management.base import OutputWrapper

# TODO: Add tests for this class


class Table:
    """A class for rendering a table on the terminal."""

    class Alignment(str, Enum):
        """An alignment of a column."""

        LEFT = "left"
        RIGHT = "right"

        @property
        def to_f(self) -> str:
            """Return as symbol for f-strings"""
            if self is self.LEFT:
                return "<"

            if self is self.RIGHT:
                return ">"

            raise NotImplementedError("Invalid alignment")

    def __init__(self, default_alignment="left") -> None:
        self._rows: List[Tuple[str, ...]] = []
        self._column_width: Tuple[int, ...] = tuple()
        self._alignment: List[Table.Alignment] = []
        self._default_alignment = self.Alignment(default_alignment)

    def set_data(self, data: List[Dict[str, Any]]):
        """Set data of this table."""
        self._rows = self._convert_to_table(data)
        self._column_widths = self._calculate_column_width()
        self._reset_alignment()

    @property
    def columns_count(self) -> int:
        """Return number of columns."""
        return len(self._rows[0])

    def _convert_to_table(self, data: List[dict]) -> List[Tuple[str, ...]]:
        table = [tuple(self._format_head(v) for v in data[0].keys())]
        table += [tuple(self._format_value(v) for v in o.values()) for o in data]
        return table

    def _format_head(self, value: str) -> str:
        return value.replace("_", " ").capitalize()

    def _format_value(self, value) -> str:
        if value is None:
            return "?"

        if isinstance(value, float):
            return f"{value:,.1f}"

        if isinstance(value, int):
            return f"{value:,}"

        return str(value)

    def _calculate_column_width(self) -> Tuple[int, ...]:
        widths = [[] for _ in self._rows[0]]
        for row in self._rows:
            for col_num, column in enumerate(row):
                widths[col_num].append(len(column))

        max_width = tuple(max(column) for column in widths)
        return max_width

    def _reset_alignment(self):
        self._alignment = [self._default_alignment for _ in range(self.columns_count)]

    def set_alignment(self, column: int, alignment: Union[Alignment, str]):
        """Set alignment for a column."""
        self._alignment[column] = self.Alignment(alignment)

    def write(self, stdout: OutputWrapper, indentation: int = 2, margin: int = 2):
        """Write table to output."""
        self._write_table(stdout, indentation, margin)

    def _write_table(self, stdout: OutputWrapper, indentation: int, margin: int):
        columns_count = self.columns_count
        for row_num, row in enumerate(self._rows):
            output_row = " " * indentation
            for col_num, column in enumerate(row):
                width = self._column_widths[col_num]
                alignment = self._alignment[col_num]
                output_row += f"{column:{alignment.to_f}{width}}"

                if col_num < columns_count - 1:
                    output_row += " " * margin

            stdout.write(output_row)

            if row_num == 0:
                output_row = " " * indentation
                for col_num, _ in enumerate(row):
                    width = self._column_widths[col_num]
                    output_row += "-" * width + " " * margin
                stdout.write(output_row)



import tempfile
from io import StringIO
from pathlib import Path

from django.core.management import call_command

from app_utils.testing import NoSocketsTestCase

from memberaudit.tests.testdata.factories import (
    create_character_contract,
    create_character_contract_item,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.utils import create_memberaudit_character

DATA_EXPORTERS_PATH = "memberaudit.core.data_exporters"


class TestDataExport(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_should_export_contract_item(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            # given
            contract = create_character_contract(character=self.character_1001)
            create_character_contract_item(contract=contract, record_id=12)
            out = StringIO()
            # when
            call_command(
                "memberaudit_data_export",
                "contract-item",
                "--destination",
                tmp_dir_name,
                stdout=out,
            )
            # then
            output_file = Path(tmp_dir_name) / Path(
                "memberaudit_contract-item"
            ).with_suffix(".csv")
            self.assertTrue(output_file.exists())



# Generated by Django 4.0.10 on 2023-12-08 23:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0015_charactership_item_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="characterupdatestatus",
            name="parent_task_id",
        ),
        migrations.RemoveField(
            model_name="characterupdatestatus",
            name="root_task_id",
        ),
        migrations.AddField(
            model_name="characterupdatestatus",
            name="update_finished_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                help_text="Finish time of an actual update",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="characterupdatestatus",
            name="update_started_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                help_text="Start time of an actual update",
                null=True,
            ),
        ),
        migrations.RenameField(
            model_name="characterupdatestatus",
            old_name="finished_at",
            new_name="run_finished_at",
        ),
        migrations.AlterField(
            model_name="characterupdatestatus",
            name="run_finished_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                help_text="Finish time of an update attempt",
                null=True,
            ),
        ),
        migrations.RenameField(
            model_name="characterupdatestatus",
            old_name="started_at",
            new_name="run_started_at",
        ),
        migrations.AlterField(
            model_name="characterupdatestatus",
            name="run_started_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                help_text="Start time of an update attempt",
                null=True,
            ),
        ),
        migrations.RenameField(
            model_name="characterupdatestatus",
            old_name="last_error_message",
            new_name="error_message",
        ),
    ]



from unittest.mock import patch

from django.http import HttpResponse
from django.test import RequestFactory, TestCase
from esi.errors import TokenError
from esi.models import Token

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.esi import EsiDailyDowntime, EsiOffline, EsiStatus
from app_utils.testing import NoSocketsTestCase, generate_invalid_pk

from memberaudit.decorators import (
    fetch_character_if_allowed,
    fetch_token_for_character,
    when_esi_is_available,
)
from memberaudit.models import Character

from .testdata.factories import create_character_from_user
from .testdata.load_entities import load_entities
from .utils import create_user_from_evecharacter_with_access, scope_names_set

MODULE_PATH = "memberaudit.decorators"

DUMMY_URL = "http://www.example.com"


class TestFetchOwnerIfAllowed(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    def test_passthrough_when_fetch_owner_if_allowed(self):
        @fetch_character_if_allowed()
        def dummy(request, character_pk, character):
            self.assertEqual(character, my_character)
            self.assertIn("eve_character", character._state.fields_cache)
            return HttpResponse("ok")

        # given
        my_character = create_character_from_user(self.user)
        user = my_character.eve_character.character_ownership.user
        request = self.factory.get(DUMMY_URL)
        request.user = user
        # when
        response = dummy(request, my_character.pk)
        # then
        self.assertEqual(response.status_code, 200)

    def test_returns_404_when_owner_not_found(self):
        @fetch_character_if_allowed()
        def dummy(request, character_pk, character):
            self.assertTrue(False)

        # given
        my_character = create_character_from_user(self.user)
        user = my_character.eve_character.character_ownership.user
        request = self.factory.get(DUMMY_URL)
        request.user = user
        # when
        response = dummy(request, generate_invalid_pk(Character))
        # then
        self.assertEqual(response.status_code, 404)

    def test_returns_403_when_user_has_not_access(self):
        @fetch_character_if_allowed()
        def dummy(request, character_pk, character):
            self.assertTrue(False)

        # given
        my_character = create_character_from_user(self.user)
        user_2 = AuthUtils.create_user("Lex Luthor")
        request = self.factory.get(DUMMY_URL)
        request.user = user_2
        # when
        response = dummy(request, my_character.pk)
        # then
        self.assertEqual(response.status_code, 403)

    # TODO: create test case with CharacterDetails
    # def test_can_specify_list_for_select_related(self):
    #     @fetch_character_if_allowed("skills")
    #     def dummy(request, character_pk, character):
    #         self.assertEqual(character, self.character)
    #         self.assertIn("skills", character._state.fields_cache)
    #         return HttpResponse("ok")

    #     OwnerSkills.objects.create(character=self.character, total_sp=10000000)
    #     request = self.factory.get(DUMMY_URL)
    #     request.user = self.user
    #     dummy(request, self.character.pk)


class TestFetchToken(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    def setUp(self) -> None:
        self.character = create_character_from_user(self.user)

    def test_defaults(self):
        @fetch_token_for_character()
        def dummy(self, character, token):
            self.assertIsInstance(token, Token)
            self.assertSetEqual(scope_names_set(token), set(Character.get_esi_scopes()))

        dummy(self, self.character)

    def test_specified_scope(self):
        @fetch_token_for_character("esi-mail.read_mail.v1")
        def dummy(self, character, token):
            self.assertIsInstance(token, Token)
            self.assertIn("esi-mail.read_mail.v1", scope_names_set(token))

        dummy(self, self.character)

    def test_exceptions_if_not_found(self):
        @fetch_token_for_character("invalid_scope")
        def dummy(self, character, token):
            pass

        with self.assertRaises(TokenError):
            dummy(self, self.character)


@when_esi_is_available
def when_esi_is_available_func():
    return "ok"


@patch(MODULE_PATH + ".IS_TESTING", False)
@patch(MODULE_PATH + ".cache.set", spec=True)
@patch(MODULE_PATH + ".cache.get", spec=True)
@patch(MODULE_PATH + ".fetch_esi_status")
class TestWhenEsiIsAvailable(NoSocketsTestCase):
    def test_should_run_task_when_esi_is_available(
        self, mock_fetch_esi_status, mock_cache_get, mock_cache_set
    ):
        # given
        mock_cache_get.return_value = None
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)
        # when/then
        result = when_esi_is_available_func()
        # then
        self.assertEqual(result, "ok")

    def test_should_raise_error_when_esi_is_offline(
        self, mock_fetch_esi_status, mock_cache_get, mock_cache_set
    ):
        # given
        mock_cache_get.return_value = None
        mock_fetch_esi_status.return_value = EsiStatus(False, 99, 60)
        # when/then
        with self.assertRaises(EsiOffline):
            when_esi_is_available_func()

    def test_should_complete_task_early_when_downtime(
        self, mock_fetch_esi_status, mock_cache_get, mock_cache_set
    ):
        # given
        mock_cache_get.return_value = None
        mock_fetch_esi_status.side_effect = EsiDailyDowntime
        # when/then
        result = when_esi_is_available_func()
        # then
        self.assertIsNone(result)

    def test_should_take_status_from_cache_when_availale(
        self, mock_fetch_esi_status, mock_cache_get, mock_cache_set
    ):
        # given
        mock_cache_get.return_value = EsiStatus(True, 99, 60)
        mock_fetch_esi_status.side_effect = RuntimeError
        # when/then
        result = when_esi_is_available_func()
        # then
        self.assertEqual(result, "ok")
        self.assertFalse(mock_cache_set.called)

    def test_should_run_task_when_esi_is_available_and_store_in_cache(
        self, mock_fetch_esi_status, mock_cache_get, mock_cache_set
    ):
        # given
        mock_cache_get.return_value = None
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)
        # when/then
        result = when_esi_is_available_func()
        # then
        self.assertEqual(result, "ok")
        self.assertTrue(mock_cache_set.called)



from unittest.mock import patch

import requests_mock

from django.test import TestCase, override_settings

from memberaudit import tasks
from memberaudit.core import esi_status
from memberaudit.tests.testdata.esi_client_stub import esi_stub
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import (
    create_memberaudit_character,
    reset_celery_once_locks,
)

MANAGERS_PATH = "memberaudit.managers"
MODELS_PATH = "memberaudit.models"
TASKS_PATH = "memberaudit.tasks"


# TODO: Replace esi_stubs with http request mocks
@patch(MANAGERS_PATH + ".character_sections_1.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_2.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_3.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_1.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_2.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_3.esi", esi_stub)
@patch(MANAGERS_PATH + ".general.esi", esi_stub)
@requests_mock.Mocker()
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestTasksIntegration(TestCase):
    fixtures = ["disable_analytics.json"]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        reset_celery_once_locks()
        esi_status.clear_cache()

    def test_should_update_all_characters(self, requests_mocker):
        # given
        character_1001 = create_memberaudit_character(1001)
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/status.json?version=latest",
            json=[
                {
                    "endpoint": "esi-mail",
                    "method": "get",
                    "route": "/characters/{character_id}/mail/",
                    "status": "green",
                    "tags": ["Mail"],
                }
            ],
        )
        requests_mocker.register_uri(
            "GET",
            url="https://esi.evetech.net/latest/status/",
            headers={
                "X-Esi-Error-Limit-Remain": "40",
                "X-Esi-Error-Limit-Reset": "30",
            },
            json={
                "players": 12345,
                "server_version": "1132976",
                "start_time": "2017-01-02T12:34:56Z",
            },
        )
        # when
        tasks.update_all_characters()
        # then
        self.assertTrue(character_1001.is_update_status_ok())



"""Determine which character sections are currently reported as unavailable by ESI."""

import dataclasses
import random
from http import HTTPStatus
from time import sleep
from typing import Any, Dict, List, Optional, Set

import requests
from requests.exceptions import RequestException

from django.core.cache import cache

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__, __version__
from memberaudit.models import Character

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

_CACHE_KEY = "memberaudit-esi-status"
_CACHE_TIMEOUT = 120
_ESI_STATUS_JSON_URL = "https://esi.evetech.net/status.json?version=latest"
_MAX_RETRIES = 3
_REQUEST_TIMEOUT = (5, 30)


@dataclasses.dataclass
class _Endpoint:
    """A ESI endpoint."""

    method: str
    route: str

    def __post_init__(self):
        if not self.method or not self.route or self.method not in {"get", "post"}:
            raise ValueError(f"invalid: {self}")

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "_Endpoint":
        """Create a new object from a dict."""
        return cls(method=d["method"], route=d["route"])


_REQUIRED_ENDPOINTS_FOR_SECTIONS = {
    Character.UpdateSection.ASSETS: [
        _Endpoint("get", "/characters/{character_id}/assets/"),
        _Endpoint("post", "/characters/{character_id}/assets/names/"),
    ],
    Character.UpdateSection.ATTRIBUTES: [
        _Endpoint("get", "/characters/{character_id}/attributes/"),
    ],
    Character.UpdateSection.CHARACTER_DETAILS: [
        _Endpoint("get", "/characters/{character_id}/"),
    ],
    Character.UpdateSection.CONTACTS: [
        _Endpoint("get", "/characters/{character_id}/contacts/"),
    ],
    Character.UpdateSection.CONTRACTS: [
        _Endpoint("get", "/characters/{character_id}/contracts/"),
        _Endpoint("get", "/characters/{character_id}/contracts/{contract_id}/bids/"),
        _Endpoint("get", "/characters/{character_id}/contracts/{contract_id}/items/"),
    ],
    Character.UpdateSection.CORPORATION_HISTORY: [
        _Endpoint("get", "/characters/{character_id}/corporationhistory/"),
    ],
    Character.UpdateSection.FW_STATS: [
        _Endpoint("get", "/characters/{character_id}/fw/stats/"),
    ],
    Character.UpdateSection.IMPLANTS: [
        _Endpoint("get", "/characters/{character_id}/implants/"),
    ],
    Character.UpdateSection.JUMP_CLONES: [
        _Endpoint("get", "/characters/{character_id}/clones/"),
    ],
    Character.UpdateSection.LOCATION: [
        _Endpoint("get", "/characters/{character_id}/location/"),
    ],
    Character.UpdateSection.LOYALTY: [
        _Endpoint("get", "/characters/{character_id}/loyalty/points/"),
    ],
    Character.UpdateSection.MAILS: [
        _Endpoint("get", "/characters/{character_id}/mail/"),
        _Endpoint("get", "/characters/{character_id}/mail/labels/"),
        _Endpoint("get", "/characters/{character_id}/mail/lists/"),
        _Endpoint("get", "/characters/{character_id}/mail/{mail_id}/"),
    ],
    Character.UpdateSection.MINING_LEDGER: [
        _Endpoint("get", "/characters/{character_id}/mining/"),
    ],
    Character.UpdateSection.ONLINE_STATUS: [
        _Endpoint("get", "/characters/{character_id}/online/"),
    ],
    Character.UpdateSection.PLANETS: [
        _Endpoint("get", "/characters/{character_id}/planets/"),
    ],
    Character.UpdateSection.ROLES: [
        _Endpoint("get", "/characters/{character_id}/roles/"),
    ],
    Character.UpdateSection.SHIP: [
        _Endpoint("get", "/characters/{character_id}/ship/"),
    ],
    Character.UpdateSection.SKILLS: [
        _Endpoint("get", "/characters/{character_id}/skills/"),
    ],
    Character.UpdateSection.SKILL_QUEUE: [
        _Endpoint("get", "/characters/{character_id}/skillqueue/"),
    ],
    Character.UpdateSection.STANDINGS: [
        _Endpoint("get", "/characters/{character_id}/standings/"),
    ],
    Character.UpdateSection.TITLES: [
        _Endpoint("get", "/characters/{character_id}/titles/"),
    ],
    Character.UpdateSection.WALLET_BALLANCE: [
        _Endpoint("get", "/characters/{character_id}/wallet/"),
    ],
    Character.UpdateSection.WALLET_JOURNAL: [
        _Endpoint("get", "/characters/{character_id}/wallet/journal/"),
    ],
    Character.UpdateSection.WALLET_TRANSACTIONS: [
        _Endpoint("get", "/characters/{character_id}/wallet/transactions/"),
    ],
}
"""Endpoints which must be available for an endpoint to function."""


def unavailable_sections() -> Optional[Set[Character.UpdateSection]]:
    """Returns a set of all sections which endpoints are currently
    reported as "red" by ESI. Returns None if there was a failure.

    An empty set means that all sections are available.

    Results are cached.
    """
    status = cache.get(_CACHE_KEY)
    if status:
        return status

    status = _unavailable_sections()
    if status is None:
        return None

    cache.set(key=_CACHE_KEY, value=status, timeout=_CACHE_TIMEOUT)
    return status


def _unavailable_sections() -> Optional[Set[Character.UpdateSection]]:
    status = _fetch_status()
    if not status:
        return None

    sections = _determine_unavailable_sections(status)
    return sections


def _fetch_status() -> Optional[List[Dict[str, Any]]]:
    try:
        r = _get_esi_status()
        r.raise_for_status()
        status = r.json()
    except RequestException as exc:
        logger.warning(f"Failed to get ESI status. Error: {exc}")
        return None
    return status


def _get_esi_status() -> requests.Response:
    """Fetch current ESI status. Retry on common HTTP errors."""
    retry_count = 0
    while True:
        response = requests.get(
            _ESI_STATUS_JSON_URL,
            timeout=_REQUEST_TIMEOUT,
            headers={"User-Agent": f"{__package__};{__version__}"},
        )
        if response.status_code not in {
            HTTPStatus.BAD_GATEWAY,
            HTTPStatus.SERVICE_UNAVAILABLE,
            HTTPStatus.GATEWAY_TIMEOUT,
        }:
            break

        retry_count += 1
        if retry_count == _MAX_RETRIES:
            break

        wait_secs = 0.1 * (random.uniform(2, 4) ** retry_count)
        logger.warning(
            "HTTP status code %s - Try %s/%s - Delay %f",
            response.status_code,
            retry_count,
            _MAX_RETRIES,
            wait_secs,
        )
        sleep(wait_secs)

    return response


def _determine_unavailable_sections(
    status: List[Dict[str, Any]],
) -> Set[Character.UpdateSection]:
    sections: Set[Character.UpdateSection] = set()
    red_endpoints = [_Endpoint.from_dict(ep) for ep in status if ep["status"] == "red"]
    for section, ep in _REQUIRED_ENDPOINTS_FOR_SECTIONS.items():
        if _is_section_broken(ep, red_endpoints):
            sections.add(section)
    return sections


def _is_section_broken(
    section_endpoints: List[_Endpoint], red_endpoints: List[_Endpoint]
) -> bool:
    for sep in section_endpoints:
        for rep in red_endpoints:
            if rep.method == sep.method and rep.route == sep.route:
                return True
    return False


def clear_cache():
    """Clear cache."""
    cache.delete(_CACHE_KEY)



# Generated by Django 4.0.10 on 2023-06-15 12:19

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("memberaudit", "0007_add_localization_2"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    (
                        "basic_access",
                        "Can access this app, register, and view own characters",
                    ),
                    ("share_characters", "Can share his/her characters"),
                    ("finder_access", "Can access character finder feature"),
                    ("reports_access", "Can access reports feature"),
                    ("characters_access", "Can view characters owned by others"),
                    ("exports_access", "Can access data exports"),
                    ("view_shared_characters", "Can view shared characters"),
                    ("view_same_corporation", "Can view corporation characters"),
                    ("view_same_alliance", "Can view alliance characters"),
                    ("view_everything", "Can view all characters"),
                    (
                        "notified_on_character_removal",
                        "Notified when member drops character",
                    ),
                    ("view_skill_sets", "Can view skill sets for a character"),
                ),
            },
        ),
        migrations.AddField(
            model_name="skillset",
            name="last_modified_at",
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="skillset",
            name="last_modified_by",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="skillsetgroup",
            name="last_modified_at",
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="skillsetgroup",
            name="last_modified_by",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



"""Character sections models."""

import datetime as dt
from typing import Optional

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity, EvePlanet, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.core import standings
from memberaudit.helpers import EveEntityIdsMixin, arabic_number_to_roman
from memberaudit.managers.character_sections_3 import (
    CharacterMiningLedgerEntryManager,
    CharacterOnlineStatusManager,
    CharacterPlanetManager,
    CharacterRoleManager,
    CharacterShipManager,
    CharacterSkillManager,
    CharacterSkillqueueEntryManager,
    CharacterSkillSetCheckManager,
    CharacterStandingManager,
    CharacterTitleManager,
    CharacterWalletBalanceManager,
    CharacterWalletJournalEntryManager,
    CharacterWalletTransactionManager,
)

from ._helpers import AddGenericReprMixin
from .characters import Character
from .constants import CURRENCY_MAX_DECIMALS, CURRENCY_MAX_DIGITS, NAMES_MAX_LENGTH
from .general import Location

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterMiningLedgerEntry(AddGenericReprMixin, models.Model):
    """Mining ledger entry of a character."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="mining_ledger"
    )
    date = models.DateField(db_index=True)
    quantity = models.PositiveIntegerField()
    eve_solar_system = models.ForeignKey(
        EveSolarSystem, on_delete=models.CASCADE, related_name="+"
    )
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    objects = CharacterMiningLedgerEntryManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "date", "eve_solar_system", "eve_type"],
                name="functional_pk_characterminingledgerentry",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character} {self.id}"


class CharacterOnlineStatus(AddGenericReprMixin, models.Model):
    """Online Status of a character."""

    character = models.OneToOneField(
        Character,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="online_status",
    )

    last_login = models.DateTimeField(default=None, null=True)
    last_logout = models.DateTimeField(default=None, null=True)
    logins = models.PositiveIntegerField(default=None, null=True)

    objects = CharacterOnlineStatusManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return str(self.character)


class CharacterPlanet(AddGenericReprMixin, models.Model):
    """A planetary colony belonging to a character."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="planets"
    )
    eve_planet = models.ForeignKey(
        EvePlanet, on_delete=models.CASCADE, related_name="+"
    )

    last_update_at = models.DateTimeField()
    num_pins = models.PositiveIntegerField()
    upgrade_level = models.PositiveIntegerField()

    objects = CharacterPlanetManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "eve_planet"], name="functional_pk_characterplanet"
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.eve_planet.name}"

    def planet_type(self) -> str:
        """Return name of the planet type."""
        return self.eve_planet.eve_type.name


class CharacterRole(AddGenericReprMixin, models.Model):
    """A character's corporation role."""

    class Location(models.TextChoices):
        """A location for a role."""

        BASE = "BS", _("base")
        HQ = "HQ", _("headquarters")
        OTHER = "OT", _("other")
        UNIVERSAL = "UV", _("universal")

    class Role(models.TextChoices):
        """A corporation role."""

        ACCOUNT_TAKE_1 = "AT1", _("account take 1")
        ACCOUNT_TAKE_2 = "AT2", _("account take 2")
        ACCOUNT_TAKE_3 = "AT3", _("account take 3")
        ACCOUNT_TAKE_4 = "AT4", _("account take 4")
        ACCOUNT_TAKE_5 = "AT5", _("account take 5")
        ACCOUNT_TAKE_6 = "AT6", _("account take 6")
        ACCOUNT_TAKE_7 = "AT7", _("account take 7")
        ACCOUNTANT = "ACT", _("accountant")
        AUDITOR = "AUD", _("auditor")
        BRAND_MANAGER = "BMG", _("brand manager")
        COMMUNICATIONS_OFFICER = "COM", _("communications officer")
        CONFIG_EQUIPMENT = "CEQ", _("config equipment")
        CONFIG_STARBASE_EQUIPMENT = "CSE", _("config starbase equipment")
        CONTAINER_TAKE_1 = "CT1", _("container take 1")
        CONTAINER_TAKE_2 = "CT2", _("container take 2")
        CONTAINER_TAKE_3 = "CT3", _("container take 3")
        CONTAINER_TAKE_4 = "CT4", _("container take 4")
        CONTAINER_TAKE_5 = "CT5", _("container take 5")
        CONTAINER_TAKE_6 = "CT6", _("container take 6")
        CONTAINER_TAKE_7 = "CT7", _("container take 7")
        CONTRACT_MANAGER = "CMG", _("contract manager")
        DELIVERIES_CONTAINER_TAKE = "DCT", _("deliveries container take")
        DELIVERIES_QUERY = "DQY", _("deliveries query")
        DELIVERIES_TAKE = "DTK", _("deliveries take")
        DIPLOMAT = "DPL", _("diplomat")
        DIRECTOR = "DRT", _("director")
        FACTORY_MANAGER = "FCM", _("factory manager")
        FITTING_MANAGER = "FTM", _("fitting manager")
        HANGAR_QUERY_1 = "HQ1", _("hangar query 1")
        HANGAR_QUERY_2 = "HQ2", _("hangar query 2")
        HANGAR_QUERY_3 = "HQ3", _("hangar query 3")
        HANGAR_QUERY_4 = "HQ4", _("hangar query 4")
        HANGAR_QUERY_5 = "HQ5", _("hangar query 5")
        HANGAR_QUERY_6 = "HQ6", _("hangar query 6")
        HANGAR_QUERY_7 = "HQ7", _("hangar query 7")
        HANGAR_TAKE_1 = "HT1", _("hangar take 1")
        HANGAR_TAKE_2 = "HT2", _("hangar take 2")
        HANGAR_TAKE_3 = "HT3", _("hangar take 3")
        HANGAR_TAKE_4 = "HT4", _("hangar take 4")
        HANGAR_TAKE_5 = "HT5", _("hangar take 5")
        HANGAR_TAKE_6 = "HT6", _("hangar take 6")
        HANGAR_TAKE_7 = "HT7", _("hangar take 7")
        JUNIOR_ACCOUNTANT = "JAC", _("junior accountant")
        PERSONNEL_MANAGER = "PSM", _("personnel manager")
        PROJECT_MANAGER = "PMG", _("project manager")
        RENT_FACTORY_FACILITY = "RFF", _("rent factory facility")
        RENT_OFFICE = "RFC", _("rent office")
        RENT_RESEARCH_FACILITY = "RRF", _("rent research facility")
        SECURITY_OFFICER = "SCO", _("security officer")
        SKILL_PLAN_MANAGER = "SPM", _("skill plan manager")
        STARBASE_DEFENSE_OPERATOR = "SDO", _("starbase defense operator")
        STARBASE_FUEL_TECHNICIAN = "SFT", _("starbase fuel technician")
        STATION_MANAGER = "STM", _("station manager")
        TRADER = "TRD", _("trader")

    _ROLES_HANGER_ACCESS = {
        Role.HANGAR_QUERY_1,
        Role.HANGAR_QUERY_2,
        Role.HANGAR_QUERY_3,
        Role.HANGAR_QUERY_4,
        Role.HANGAR_QUERY_5,
        Role.HANGAR_QUERY_6,
        Role.HANGAR_QUERY_7,
        Role.HANGAR_TAKE_1,
        Role.HANGAR_TAKE_2,
        Role.HANGAR_TAKE_3,
        Role.HANGAR_TAKE_4,
        Role.HANGAR_TAKE_5,
        Role.HANGAR_TAKE_6,
        Role.HANGAR_TAKE_7,
    }
    _ROLES_CONTAINER_ACCESS = {
        Role.CONTAINER_TAKE_1,
        Role.CONTAINER_TAKE_2,
        Role.CONTAINER_TAKE_3,
        Role.CONTAINER_TAKE_4,
        Role.CONTAINER_TAKE_5,
        Role.CONTAINER_TAKE_6,
        Role.CONTAINER_TAKE_7,
    }
    _ROLES_DELIVERIES_ACCESS = {
        Role.DELIVERIES_CONTAINER_TAKE,
        Role.DELIVERIES_QUERY,
        Role.DELIVERIES_TAKE,
    }
    ROLES_GROUPED = [
        {
            "title": _("general roles"),
            "location": Location.UNIVERSAL,
            "roles": {
                Role.ACCOUNTANT,
                Role.AUDITOR,
                Role.BRAND_MANAGER,
                Role.COMMUNICATIONS_OFFICER,
                Role.CONFIG_EQUIPMENT,
                Role.CONFIG_STARBASE_EQUIPMENT,
                Role.CONTRACT_MANAGER,
                Role.DIPLOMAT,
                Role.DIRECTOR,
                Role.FITTING_MANAGER,
                Role.JUNIOR_ACCOUNTANT,
                Role.PERSONNEL_MANAGER,
                Role.PROJECT_MANAGER,
                Role.STARBASE_DEFENSE_OPERATOR,
                Role.STARBASE_FUEL_TECHNICIAN,
            },
        },
        {
            "title": _("station services"),
            "location": Location.UNIVERSAL,
            "roles": {
                Role.FACTORY_MANAGER,
                Role.RENT_FACTORY_FACILITY,
                Role.RENT_OFFICE,
                Role.RENT_RESEARCH_FACILITY,
                Role.SECURITY_OFFICER,
                Role.STATION_MANAGER,
                Role.TRADER,
            },
        },
        {
            "title": _("accounting (divisional)"),
            "location": Location.UNIVERSAL,
            "roles": {
                Role.ACCOUNT_TAKE_1,
                Role.ACCOUNT_TAKE_2,
                Role.ACCOUNT_TAKE_3,
                Role.ACCOUNT_TAKE_4,
                Role.ACCOUNT_TAKE_5,
                Role.ACCOUNT_TAKE_6,
                Role.ACCOUNT_TAKE_7,
            },
        },
        {
            "title": _("hangar access (headquarters)"),
            "location": Location.HQ,
            "roles": _ROLES_HANGER_ACCESS,
        },
        {
            "title": _("container access (headquarters)"),
            "location": Location.HQ,
            "roles": _ROLES_CONTAINER_ACCESS,
        },
        {
            "title": _("deliveries (headquarters)"),
            "location": Location.HQ,
            "roles": _ROLES_DELIVERIES_ACCESS,
        },
        {
            "title": _("hangar access (based at)"),
            "location": Location.BASE,
            "roles": _ROLES_HANGER_ACCESS,
        },
        {
            "title": _("deliveries (based at)"),
            "location": Location.BASE,
            "roles": _ROLES_DELIVERIES_ACCESS,
        },
        {
            "title": _("container access (based at)"),
            "location": Location.BASE,
            "roles": _ROLES_CONTAINER_ACCESS,
        },
        {
            "title": _("hangar access (other)"),
            "location": Location.OTHER,
            "roles": _ROLES_HANGER_ACCESS,
        },
        {
            "title": _("deliveries (other)"),
            "location": Location.OTHER,
            "roles": _ROLES_DELIVERIES_ACCESS,
        },
        {
            "title": _("container access (other)"),
            "location": Location.OTHER,
            "roles": _ROLES_CONTAINER_ACCESS,
        },
    ]

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="roles"
    )

    location = models.CharField(
        max_length=2,
        choices=Location.choices,
        help_text=_("Location where this role is applicable."),
    )
    role = models.CharField(
        max_length=3, choices=Role.choices, help_text=_("Role of a character")
    )

    objects = CharacterRoleManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "location", "role"],
                name="functional_pk_characterrole",
            )
        ]


class CharacterShip(AddGenericReprMixin, models.Model):
    """The current ship of a character."""

    character = models.OneToOneField(
        Character, on_delete=models.CASCADE, related_name="ship"
    )

    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    item_id = models.PositiveBigIntegerField()
    name = models.CharField(max_length=255)

    objects = CharacterShipManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return str(f"{self.character}-{self.eve_type.name}")


class CharacterSkill(AddGenericReprMixin, models.Model):
    """A trained skill of a character."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="skills"
    )
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    active_skill_level = models.PositiveIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)]
    )
    skillpoints_in_skill = models.PositiveBigIntegerField()
    trained_skill_level = models.PositiveBigIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)]
    )

    objects = CharacterSkillManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "eve_type"], name="functional_pk_characterskill"
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.eve_type.name}"


class CharacterSkillpoints(AddGenericReprMixin, models.Model):
    """The skill points of a character."""

    character = models.OneToOneField(
        Character,
        primary_key=True,
        on_delete=models.CASCADE,
        related_name="skillpoints",
    )
    total = models.PositiveBigIntegerField()
    unallocated = models.PositiveIntegerField(default=None, null=True)

    class Meta:
        default_permissions = ()

    def __str__(self):
        return str(self.character)


class CharacterSkillqueueEntry(AddGenericReprMixin, models.Model):
    """An entry in the skillqueue of a character."""

    character = models.ForeignKey(
        Character,
        on_delete=models.CASCADE,
        related_name="skillqueue",
    )
    queue_position = models.PositiveIntegerField(db_index=True)

    finish_date = models.DateTimeField(default=None, null=True)
    finished_level = models.PositiveIntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(5)]
    )
    level_end_sp = models.PositiveIntegerField(default=None, null=True)
    level_start_sp = models.PositiveIntegerField(default=None, null=True)
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    start_date = models.DateTimeField(default=None, null=True)
    training_start_sp = models.PositiveIntegerField(default=None, null=True)

    objects = CharacterSkillqueueEntryManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "queue_position"],
                name="functional_pk_characterskillqueueentry",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.skill_display()}"

    def completion_percent(self) -> float:
        """Return current training progress for a skill."""
        duration = self.total_duration()
        if duration is None:
            raise ValueError("insufficient data to calculate")

        now_ = now()
        if self.finish_date < now_:
            return 1
        if self.start_date > now_:
            return 0
        # if duration.total_seconds() == 0:
        #     return 1
        if (
            self.level_start_sp is None
            or self.level_end_sp is None
            or self.training_start_sp is None
        ):
            raise ValueError("insufficient data to calculate")
        remaining = self.finish_date - now_
        c = remaining.total_seconds() / duration.total_seconds()
        base = (self.level_end_sp - self.training_start_sp) / (
            self.level_end_sp - self.level_start_sp
        )
        return 1 - (c * base)

    def is_active(self) -> bool:
        """Reports whether a skill is currently being trained."""
        now_ = now()
        return (
            bool(self.start_date)
            and bool(self.finish_date)
            and self.start_date < now_ < self.finish_date
        )

    def is_completed(self) -> bool:
        """Reports whether a skill has completed training."""
        return self.completion_percent() == 1

    def remaining_duration(self) -> Optional[dt.timedelta]:
        """Return remaining duration to train a skill."""
        if not self.start_date or not self.finish_date:
            return None
        duration = self.total_duration()
        if duration is None:
            return None
        remaining_percent = 1 - self.completion_percent()
        return duration * remaining_percent

    def skill_display(self) -> str:
        """Return skill for display."""
        level_roman = arabic_number_to_roman(self.finished_level)
        name = f"{self.eve_type.name} {level_roman}"
        if not self.is_active():
            return name
        completion = self.completion_percent() * 100
        name += f" ({completion:.0f}%)"
        return name

    def total_duration(self) -> Optional[dt.timedelta]:
        """Return duration from start to finish for training a skill."""
        if not self.start_date or not self.finish_date:
            return None
        return self.finish_date - self.start_date


class CharacterSkillSetCheck(AddGenericReprMixin, models.Model):
    """The result of a skill check of a character against a skill set."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="skill_set_checks"
    )
    skill_set = models.ForeignKey("SkillSet", on_delete=models.CASCADE)

    failed_required_skills = models.ManyToManyField(
        "SkillSetSkill", related_name="failed_required_skill_set_checks"
    )
    failed_recommended_skills = models.ManyToManyField(
        "SkillSetSkill", related_name="failed_recommended_skill_set_checks"
    )

    objects = CharacterSkillSetCheckManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "skill_set"],
                name="functional_pk_characterskillsetcheck",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.skill_set}"

    @property
    def can_fly(self) -> bool:
        """Return True if character has all required skills for this skill set,
        else False.
        """
        return not self.failed_required_skills.exists()


class CharacterStanding(AddGenericReprMixin, models.Model):
    """Standing of a character with an NPC entity in Eve Online."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="standings"
    )
    eve_entity = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="+"
    )

    standing = models.FloatField()

    objects = CharacterStandingManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "eve_entity"],
                name="functional_pk_characterstanding",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.eve_entity}"

    def effective_standing(
        self,
        connections_skill_level: int,
        criminal_connections_skill_level: int,  # pylint: disable=unused-argument
        diplomacy_skill_level: int,
    ) -> float:
        """Return effective standing for this NPC after applying social skill."""
        unadjusted_standing = self.standing
        if unadjusted_standing >= 0:
            skill_level = connections_skill_level
            skill_modifier = 0.04
            # TODO: Add variant for criminal connection
        else:
            skill_level = diplomacy_skill_level
            skill_modifier = 0.04

        max_possible_standing = 10
        effective_standing = standings.calc_effective_standing(
            unadjusted_standing, skill_level, skill_modifier, max_possible_standing
        )
        return effective_standing


class CharacterTitle(AddGenericReprMixin, models.Model):
    """Title of a character."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="titles"
    )
    title_id = models.PositiveIntegerField()

    name = models.CharField(max_length=NAMES_MAX_LENGTH)

    objects = CharacterTitleManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "title_id"],
                name="functional_pk_charactertitle",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.name}"


class CharacterWalletBalance(AddGenericReprMixin, models.Model):
    """A wallet balance of a character."""

    character = models.OneToOneField(
        Character,
        primary_key=True,
        on_delete=models.CASCADE,
        related_name="wallet_balance",
    )
    total = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS, decimal_places=CURRENCY_MAX_DECIMALS
    )

    objects = CharacterWalletBalanceManager()

    class Meta:
        default_permissions = ()

    def __str__(self):
        return str(self.character)


class CharacterWalletJournalEntry(EveEntityIdsMixin, AddGenericReprMixin, models.Model):
    """A wallet journal entry of a character in Eve Online."""

    CONTEXT_ID_TYPE_UNDEFINED = "NON"
    CONTEXT_ID_TYPE_STRUCTURE_ID = "STR"
    CONTEXT_ID_TYPE_STATION_ID = "STA"
    CONTEXT_ID_TYPE_MARKET_TRANSACTION_ID = "MTR"
    CONTEXT_ID_TYPE_CHARACTER_ID = "CHR"
    CONTEXT_ID_TYPE_CORPORATION_ID = "COR"
    CONTEXT_ID_TYPE_ALLIANCE_ID = "ALL"
    CONTEXT_ID_TYPE_EVE_SYSTEM = "EVE"
    CONTEXT_ID_TYPE_INDUSTRY_JOB_ID = "INJ"
    CONTEXT_ID_TYPE_CONTRACT_ID = "CNT"
    CONTEXT_ID_TYPE_PLANET_ID = "PLN"
    CONTEXT_ID_TYPE_SYSTEM_ID = "SYS"
    CONTEXT_ID_TYPE_TYPE_ID = "TYP"
    CONTEXT_ID_CHOICES = (
        (CONTEXT_ID_TYPE_UNDEFINED, _("undefined")),
        (CONTEXT_ID_TYPE_STATION_ID, _("station ID")),
        (CONTEXT_ID_TYPE_MARKET_TRANSACTION_ID, _("market transaction ID")),
        (CONTEXT_ID_TYPE_CHARACTER_ID, _("character ID")),
        (CONTEXT_ID_TYPE_CORPORATION_ID, _("corporation ID")),
        (CONTEXT_ID_TYPE_ALLIANCE_ID, _("alliance ID")),
        (CONTEXT_ID_TYPE_EVE_SYSTEM, _("eve system")),
        (CONTEXT_ID_TYPE_INDUSTRY_JOB_ID, _("industry job ID")),
        (CONTEXT_ID_TYPE_CONTRACT_ID, _("contract ID")),
        (CONTEXT_ID_TYPE_PLANET_ID, _("planet ID")),
        (CONTEXT_ID_TYPE_SYSTEM_ID, _("system ID")),
        (CONTEXT_ID_TYPE_TYPE_ID, _("type ID")),
    )
    CONTEXT_ID_MAPS = {
        "undefined": CONTEXT_ID_TYPE_UNDEFINED,
        "station_id": CONTEXT_ID_TYPE_STATION_ID,
        "market_transaction_id": CONTEXT_ID_TYPE_MARKET_TRANSACTION_ID,
        "character_id": CONTEXT_ID_TYPE_CHARACTER_ID,
        "corporation_id": CONTEXT_ID_TYPE_CORPORATION_ID,
        "alliance_id": CONTEXT_ID_TYPE_ALLIANCE_ID,
        "eve_system": CONTEXT_ID_TYPE_EVE_SYSTEM,
        "industry_job_id": CONTEXT_ID_TYPE_INDUSTRY_JOB_ID,
        "contract_id": CONTEXT_ID_TYPE_CONTRACT_ID,
        "planet_id": CONTEXT_ID_TYPE_PLANET_ID,
        "system_id": CONTEXT_ID_TYPE_SYSTEM_ID,
        "type_id": CONTEXT_ID_TYPE_TYPE_ID,
    }

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="wallet_journal"
    )
    entry_id = models.PositiveBigIntegerField(db_index=True)

    amount = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
        blank=True,
    )
    balance = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
        blank=True,
    )
    context_id = models.PositiveBigIntegerField(default=None, null=True)
    context_id_type = models.CharField(max_length=3, choices=CONTEXT_ID_CHOICES)
    date = models.DateTimeField()
    description = models.TextField()
    first_party = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    reason = models.TextField()
    ref_type = models.CharField(max_length=64)
    second_party = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    tax = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
        blank=True,
    )
    tax_receiver = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )

    objects = CharacterWalletJournalEntryManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "entry_id"],
                name="functional_pk_characterwalletjournalentry",
            )
        ]

    def __str__(self) -> str:
        return str(self.character) + " " + str(self.entry_id)

    @classmethod
    def match_context_type_id(cls, query: str) -> str:
        """Return context type matching the query string, else UNDEFINED."""
        result = cls.CONTEXT_ID_MAPS.get(query)
        if result:
            return result

        return cls.CONTEXT_ID_TYPE_UNDEFINED


class CharacterWalletTransaction(EveEntityIdsMixin, AddGenericReprMixin, models.Model):
    """A wallet transaction of a character in Eve Online."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="wallet_transactions"
    )
    transaction_id = models.PositiveBigIntegerField(db_index=True)

    client = models.ForeignKey(EveEntity, on_delete=models.CASCADE, related_name="+")
    date = models.DateTimeField()
    is_buy = models.BooleanField()
    is_personal = models.BooleanField()
    journal_ref = models.OneToOneField(
        CharacterWalletJournalEntry,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        related_name="wallet_transaction",
    )
    location = models.ForeignKey(Location, on_delete=models.CASCADE)
    quantity = models.PositiveIntegerField()
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE)
    unit_price = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS, decimal_places=CURRENCY_MAX_DECIMALS
    )

    objects = CharacterWalletTransactionManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "transaction_id"],
                name="functional_pk_characterwallettransactions",
            )
        ]

    def __str__(self) -> str:
        return str(self.character) + " " + str(self.transaction_id)



"""Eve Online Skills."""

from collections import defaultdict
from dataclasses import dataclass
from typing import DefaultDict, Dict, Iterable, List, Tuple

from eveuniverse.models import EveType, EveTypeDogmaAttribute

from memberaudit.constants import EveDogmaAttributeId as AttributeId
from memberaudit.helpers import arabic_number_to_roman


@dataclass(frozen=True)
class Skill:
    """A skill in Eve Online."""

    eve_type: EveType
    level: int

    def __str__(self) -> str:
        level_str = arabic_number_to_roman(self.level)
        return f"{self.eve_type.name} {level_str}"

    def __lt__(self, other):
        if self.eve_type != other.eve_type:
            raise ValueError("'<' not supported for skills of different type")
        return self.level < other.level

    def __le__(self, other):
        if self.eve_type != other.eve_type:
            raise ValueError("'<=' not supported for skills of different type")
        return self.level <= other.level

    def __gt__(self, other):
        if self.eve_type != other.eve_type:
            raise ValueError("'>' not supported for skills of different type")
        return self.level > other.level

    def __ge__(self, other):
        if self.eve_type != other.eve_type:
            raise ValueError("'>=' not supported for skills of different type")
        return self.level >= other.level


def compress_skills(skills: List["Skill"]) -> List["Skill"]:
    """Compresses a list of skill by removing redundant skills."""
    skills_map = defaultdict(list)
    for skill in skills:
        skills_map[skill.eve_type.id].append(skill)
    return [max(same_skills) for _, same_skills in skills_map.items()]


def required_skills_from_eve_types(
    eve_types: Iterable[EveType],
) -> List["Skill"]:
    """Create list of required skills from eve types.

    For best performance make sure that all types have been loaded with dogmas.
    """
    skills_raw = _identify_skills_from_eve_types(eve_types)
    skill_types = _gather_skill_types(skills_raw)
    skills = [
        Skill(eve_type=skill_types[type_id], level=level)
        for type_id, level in skills_raw
    ]
    return sorted(skills, key=lambda o: o.eve_type.name)


def _identify_skills_from_eve_types(
    eve_types: Iterable[EveType],
) -> List[Tuple[int, int]]:
    eve_types = _reload_eve_types_without_dogmas(eve_types)
    all_attributes_map = _fetch_attributes_for_eve_types(eve_types)
    return _create_skills_from_attributes(all_attributes_map)


def _reload_eve_types_without_dogmas(eve_types: Iterable[EveType]) -> Iterable[EveType]:
    for eve_type in eve_types:
        if not eve_type.enabled_sections.dogmas:
            eve_type, _ = EveType.objects.update_or_create_esi(
                id=eve_type.id, enabled_sections=[EveType.Section.DOGMAS]
            )
    return eve_types


def _fetch_attributes_for_eve_types(
    eve_types: Iterable[EveType],
) -> Dict[int, Dict[int, int]]:
    eve_type_ids = {obj.id for obj in eve_types}
    all_attributes_raw = EveTypeDogmaAttribute.objects.filter(
        eve_dogma_attribute_id__in=[
            AttributeId.REQUIRED_SKILL_1,
            AttributeId.REQUIRED_SKILL_1_LEVEL,
            AttributeId.REQUIRED_SKILL_2,
            AttributeId.REQUIRED_SKILL_2_LEVEL,
            AttributeId.REQUIRED_SKILL_3,
            AttributeId.REQUIRED_SKILL_3_LEVEL,
            AttributeId.REQUIRED_SKILL_4,
            AttributeId.REQUIRED_SKILL_4_LEVEL,
            AttributeId.REQUIRED_SKILL_5,
            AttributeId.REQUIRED_SKILL_5_LEVEL,
            AttributeId.REQUIRED_SKILL_6,
            AttributeId.REQUIRED_SKILL_6_LEVEL,
        ],
        eve_type_id__in=eve_type_ids,
    ).values_list("eve_type_id", "eve_dogma_attribute_id", "value")
    all_attributes_map: DefaultDict[int, Dict[int, int]]
    all_attributes_map = defaultdict(dict)
    for eve_type_id, eve_dogma_attribute_id, value in all_attributes_raw:
        all_attributes_map[eve_type_id][eve_dogma_attribute_id] = int(value)
    return all_attributes_map


def _create_skills_from_attributes(
    all_attributes_map: Dict[int, Dict[int, int]],
) -> List[Tuple[int, int]]:
    skills = []
    for attributes in all_attributes_map.values():
        for skill_id, skill_level_id in {
            AttributeId.REQUIRED_SKILL_1: AttributeId.REQUIRED_SKILL_1_LEVEL,
            AttributeId.REQUIRED_SKILL_2: AttributeId.REQUIRED_SKILL_2_LEVEL,
            AttributeId.REQUIRED_SKILL_3: AttributeId.REQUIRED_SKILL_3_LEVEL,
            AttributeId.REQUIRED_SKILL_4: AttributeId.REQUIRED_SKILL_4_LEVEL,
            AttributeId.REQUIRED_SKILL_5: AttributeId.REQUIRED_SKILL_5_LEVEL,
            AttributeId.REQUIRED_SKILL_6: AttributeId.REQUIRED_SKILL_6_LEVEL,
        }.items():
            if skill_id in attributes and skill_level_id in attributes:
                skills.append((attributes[skill_id], attributes[skill_level_id]))
    return skills


def _gather_skill_types(skills_raw: List[Tuple[int, int]]) -> Dict[int, EveType]:
    type_ids = {skill[0] for skill in skills_raw}
    skill_types = {obj.id: obj for obj in EveType.objects.filter(id__in=type_ids)}
    missing_type_ids = type_ids - set(skill_types.keys())
    for type_id in missing_type_ids:
        eve_type, _ = EveType.objects.get_or_create_esi(id=type_id)
        skill_types[type_id] = eve_type
    return skill_types
